; Dunjunz - Games Master Edition
; Copyright (c) 2017-2020 Ciaran Anscomb
;
; See COPYING file for redistribution conditions.

; This finishes off Bank#0.  Assembled last.  Unpacks RAM-based portion of
; game.

; ========================================================================

		include "dragonhw.s"
		include "defs.s"
		include "macros.s"

		include "bank0a.exp"
		include "bank2.exp"
		include "bank3.exp"
		include "dunjunz.exp"

; ========================================================================

		org INIT_exec
		setdp 0		; assumed

		orcc #$50
		lds #$0100

		lda #gamedp
		tfr a,dp
		setdp gamedp

		lda #$e2
		sta vdg_mode

		ldd #DATA_end
		std dzip_end
		ldx #dunjunz_bin_dz
		ldu #CODE_start
		jsr dunzip

		; patch reset vector
		ldd #init_reset
		std $0072
		lda #$55
		sta $0071

init_reset
		nop
		orcc #$50
		lds #$0100

		lda #$ff
		tfr a,dp
		setdp $ff

		ldx #fb0
		ldd #$aaaa
10		std ,x++
		cmpx #fb0_end
		blo 10B

		ldd #$343c
		sta reg_pia0_cra	; HS disabled
		sta reg_pia1_cra	; printer FIRQ disabled
		stb reg_pia1_crb	; CART FIRQ disabled
		inca
		sta reg_pia0_crb	; FS enabled hi->lo

		; SAM MPU rate = slow
		sta reg_sam_r1c
		sta reg_sam_r0c

		; SAM display offset = $0200
		sta reg_sam_f0s
		sta reg_sam_f1c
		sta reg_sam_f2c

		; SAM VDG mode = G6R,G6C
		sta reg_sam_v0c
		sta reg_sam_v1s
		sta reg_sam_v2s

		lda #gamedp
		tfr a,dp
		setdp gamedp

		ldd #$aa00
		std colour_bg
		ldd #$55ff
		std colour_hi
		stb dead
		stb video_mode

		; bit of a kludge - always call the PAL initialisation
		; so we don't have to worry about updating pointers if
		; things get patched for CoCo 3
		jsr advance_video_mode

		lda $a000		; [$A000] points to ROM1 in CoCos
		anda #$20
		beq 10F			; Dragon - no patching necessary
		ldu #coco_patchset	; Apply CoCo patches
		jsr apply_patchset
		; U now points to coco3_patchset
		lda $80fd
		cmpa #'2
		bne 10F
		jsr apply_patchset
		; more kludge - *always* advance video mode at least once
		; (to RGB), but do it again if CMP detected.  this ensures
		; palette values are initialised either way.
		jsr advance_video_mode
		lda $ffb3		; test "red" palette entry
		anda #$3f
		cmpa #$24		; default from RGB palette?
		bne 25F			; nope - CMP - advance mode
		bra 30F			; yep - RGB - skip advance mode
10

		; NTSC detection
		ldx #0
		lda reg_pia0_pdrb	; clear outstanding IRQ
		sync			; wait for FS hi->lo
		lda reg_pia0_pdrb	; clear outstanding IRQ
10		lda reg_pia0_crb	; 5 - test for IRQ
		bmi 20F			; 3
		mul			; 11 - waste time...
		mul			; 11
		mul			; 11
		exg a,a			; 8
		leax 1,x		; 5 - inc line count
		bra 10B			; 3 = 57 (cycles per scanline)
20		cmpx #288		; PAL = 312, NTSC = 262
		bhs 30F
25		jsr advance_video_mode
30

		; patch reset vector
		ldd #game_reset
		std $0072
		ldd #$55ff
		sta $0071
		; set level = -1
		stb level

		clr difficulty
		clr playing

		jmp first_init_players

; CoCo patches

coco_patchset

		PATCH tbl_kdsp_menu
		KEYDSP_ENTRY 7,3,start_game	; SPACE
		KEYDSP_ENTRY 1,4,toggle_p1	; '1'
		KEYDSP_ENTRY 2,4,toggle_p2	; '2'
		KEYDSP_ENTRY 3,4,toggle_p3	; '3'
		KEYDSP_ENTRY 4,4,toggle_p4	; '4'
		KEYDSP_ENTRY 4,0,toggle_skill	; 'D'
		KEYDSP_ENTRY 6,2,toggle_vmode	; 'V'
		ENDPATCH

		PATCH tbl_p1_keys
		KEYTBL_ENTRY 1,4,2			; '1' = magic
		KEYTBL_ENTRY 2,4,$80|facing_up		; '2' = up
		KEYTBL_ENTRY 7,2,$80|facing_down	; 'W' = down
		KEYTBL_ENTRY 4,4,$80|facing_right	; '4' = right
		KEYTBL_ENTRY 3,4,$80|facing_left	; '3' = left
		KEYTBL_ENTRY 5,4,1			; '5' = fire
		fdb 0
		KEYTBL_ENTRY 7,5,2			; '/' = magic
		KEYTBL_ENTRY 0,2,$80|facing_up		; 'P' = up
		KEYTBL_ENTRY 3,5,$80|facing_down	; ';' = down
		KEYTBL_ENTRY 6,5,$80|facing_right	; '.' = right
		KEYTBL_ENTRY 4,5,$80|facing_left	; ',' = left
		KEYTBL_ENTRY 0,4,1			; '0' = fire
		fdb 0
		ENDPATCH

		PATCH PATCH_text_p1_controls_ptr+2
		fdb text_p1_controls_coco
		ENDPATCH
		PATCH PATCH_text_p2_controls_ptr+2
		fdb text_p2_controls_coco
		ENDPATCH

		PATCH PATCH_p1ddrb
		fcb $f8		; lda #$f8 - VDG ONLY as outputs
		ENDPATCH

		ENDPATCHSET

coco3_patchset

		PATCH PATCH_screen_border
		fdb $aaaa	; "blue" (remapped to black)
		ENDPATCH

		PATCH PATCH_max_video_mode
		fcb 2		; only CMP and RGB
		ENDPATCH

		PATCH tbl_text_vmode
		fdb text_cc3_cmp
		fdb text_cc3_rgb
		ENDPATCH

		PATCH video_mode_transitions+2
		fdb cc3_rgb_cmp_patchset
		fdb tiles_pal
		fdb cc3_cmp_rgb_patchset
		ENDPATCH

		ENDPATCHSET

dunjunz_bin_dz	includebin "dunjunz.bin.dz"
dunjunz_bin_dz_end

; ========================================================================

bank0_end

		align $4000,$ff
