; Dunjunz - Games Master Edition
; Copyright (c) 2017-2020 Ciaran Anscomb
;
; See COPYING file for redistribution conditions.

; This file contains common symbols, zero page allocations and the BSS.

; ========================================================================

; Definitions

DEBUG_LEVEL25	equ 0

fb0		equ $0200	; framebuffer base
fb_w		equ 32		; framebuffer line width
fb_h		equ 192		; framebuffer height
fb0_end		equ fb0+fb_w*fb_h

; Tile dimensions in pixels.  Information only.  Check 'yx_to_woff' for
; how (y,x) is actually mapped to window offsets.

tile_w		equ 12
tile_h		equ 9

; Size of wall and pickup bitmaps.

bmp_w		equ 4
bmp_h		equ 48
sizeof_bmp	equ bmp_w*bmp_h

; Window base addresses

p1_win		equ fb0+20*fb_w+1
p2_win		equ fb0+20*fb_w+14
p3_win		equ fb0+100*fb_w+1
p4_win		equ fb0+100*fb_w+14

; Stats window base addresses

p1_swin		equ fb0+21*fb_w+27
p2_swin		equ p1_swin+40*fb_w
p3_swin		equ p1_swin+80*fb_w
p4_swin		equ p1_swin+120*fb_w

; Player window dimensions in tiles.  Here for information only: code
; assumes these values and they must not change.

win_w		equ 8
win_h		equ 8

; Maximum numbers of certain types of object.

max_doors	equ 20
max_keys	equ 20
max_items	equ 30		; including exit
max_trapdoors	equ 8		; *must* be this amount of trapdoors

max_monsters	equ 28		; needn't all be active at once

; Players, monsters, shots: need to know which way each of them is facing.
; Multiples of 2 for indexing into jump tables.  Order important as bit
; twiddling is used to reverse the direction of shots.

facing_up	equ 0
facing_down	equ 2
facing_left	equ 4
facing_right	equ 6

; ========================================================================

; Structures
; ----------

; _obj_ - object.  Type is indicated by 'tile'.  Reserves a few bytes of
; arbitrary data, use of which is type-dependent.

		org 0
obj_y		rmb 1		; 'y' coordinate in map (-$80 to $3f)
obj_x		rmb 1		; 'x' coordinate in map (0 to 7)
obj_tile	rmb 1		; tile id
obj_woff	rmb 2		; precalc offset into window
obj_data	rmb 3
sizeof_obj

obj_drainer_hp	equ obj_data
obj_key_opens	equ obj_data
obj_shot_facing	equ obj_data
obj_shot_plr	equ obj_data+1

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; _spr_ - sprite.  Extends _obj_.

spr_org_	org 0-spr_offset_

spr_next_y	rmb 1		; destination 'y' when moving
spr_next_x	rmb 1		; destination 'x' when moving

spr_hp		rmb 1		; hit points (BCD)
spr_state	rmb 1		; state machine index

spr_cx		rmb 1		; current 'x' - see note

spr_offset_	equ *-spr_org_
		rmb sizeof_obj

spr_tilebase	rmb 1		; lowest tile id for player tiles

sizeof_sprite	equ *+spr_offset_

spr_facing	equ obj_data	; one of facing_*
spr_owoff	equ obj_data+1	; origin win offset (undraw during move)
spr_tmp0	equ obj_data+1	; use as a counter when not moving

; Note: spr_cx is only used for its lowest bit, which is modified during
; horizontal movement and changes which tileset to use when drawing.

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; _mon_ - monster.  Inherits from _spr_.

mon_org_	org 0-mon_offset_

mon_offset_	equ *-mon_org_+spr_offset_
		rmb sizeof_sprite

sizeof_mon	equ *+mon_offset_

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; _plr_ - player.  Extends _spr_.

plr_org_	org 0-plr_offset_

; 'control' - populated while reading keyboard/joysticks.
;   $00 = no action
;   $01 = fire
;   $02 = magic
;   $80|facing = move

plr_control	rmb 1

plr_nshots	rmb 1		; number of active shots (<= ammo)
plr_key_opens	rmb 2		; (y,x) of door that held key opens
plr_speed	rmb 1		; non-zero = player moves fast!
plr_score2	rmb 1		; BCD
plr_score0	rmb 1		; BCD
plr_yroom	rmb 1		; obj_y & 0xf8
plr_bit		rmb 1		; bit within playing/done

plr_offset_	equ *-plr_org_+spr_offset_
		rmb sizeof_sprite

plr_swin	rmb 2		; address of stats window
plr_win		rmb 2		; address of player window
plr_armour	rmb 1		; armour
plr_power	rmb 1		; weapon power
plr_ammo	rmb 1		; ammunition (1 to 3)

sizeof_plr	equ *+plr_offset_

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; _lvl_ - level source data.

		org 0
lvl_doors	rmb max_doors*2		; y,x
lvl_keys	rmb max_keys*2		; y,x
lvl_items	rmb max_items*3		; tile,y,x
lvl_trapdoors	rmb max_trapdoors*2	; y,x
lvl_bmp_wall	rmb sizeof_bmp		; wall bitmap
sizeof_lvl

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; _psg_ - PSG data

		org 0
psg_ptr		rmb 2		; tone data pointer
psg_counter	rmb 1		; tone counter
psg_eptr	rmb 2		; envelope data pointer
psg_ecounter	rmb 1		; envelope counter
psg_atten	rmb 1		; current attenuation (0 to 127)
psg_arate	rmb 1		; attenuation rate of change (-127 to 127)
psg_areg	rmb 1		; ORed into attenuation commands
sizeof_psg

; ========================================================================

; Direct page variables

; Trying to keep player sprites here for now, as there are plenty of
; places where we manipulate them directly rather than as offsets into the
; player struct.

		org $0112
gamedp		equ *>>8

tmp0		rmb 2
dzip_end
tmp1		rmb 2
tmp2		rmb 2

playing		rmb 1		; bits 0-3 = player 1-4 active
finished	rmb 1		; bits 0-3 = player 1-4 finished level
dying		rmb 1		; bits 0-3 = player 1-4 dying or dead
dead		rmb 1		; bits 0-3 = player 1-4 dead
exiting		rmb 1		; non-zero = a player is using exit

; 'death_fifo' is a list of pointers to _plr_ struct in order of player
; death.  'death_fifo_end' points to end of list, where next dead player
; is stored.  FIFO is shifted on resurrection.

death_fifo_end	rmb 2		; ptr to end of death fifo
death_fifo	rmb 8

finish_delay	rmb 1

video_mode	rmb 1		; 0=pal, 1=ntsc phase0, 2=ntsc phase1

; Font rendering
text_descender	rmb 1
text_char_width	rmb 1
text_line_x	rmb 1
text_cursor_y	rmb 1
text_cursor_x	rmb 1
text_colour	rmb 1
text_tmp0	rmb 2
text_tmp1	rmb 2

colour_bg	rmb 1
colour_fg	rmb 1
colour_hi	rmb 1
colour_lo	rmb 1

; PSG timers & pointers
psg_c1		rmb sizeof_psg
psg_c2		rmb sizeof_psg
psg_c3		rmb sizeof_psg

psg_latch_c1	rmb 2
psg_latch_c2	rmb 2
psg_latch_c3	rmb 2

difficulty	rmb 1
level		rmb 1

anim_count	rmb 1

p3_moved	rmb 1
p4_moved	rmb 1

; Player structs

players		equ *+plr_offset_
player1		equ *+plr_offset_
		rmb sizeof_plr
player2		equ *+plr_offset_
		rmb sizeof_plr
player3		equ *+plr_offset_
		rmb sizeof_plr
player4		equ *+plr_offset_
		rmb sizeof_plr
players_end	equ *+plr_offset_

; ========================================================================

BSS_start_	BSS

		org fb0_end

monsters	equ *+mon_offset_
		rmb max_monsters*sizeof_mon
monsters_end	equ *+mon_offset_

; Wall bitmap.  Bits set where wall or door exists.  Tested against before
; moving, and door bits cleared when opened.

bmp_wall	rmb sizeof_bmp
bmp_wall_end

; Objects bitmap.  Bits set where any object lies.  Used to determine if
; pickups need processed, or if a non-floor tile needs redrawn when a
; sprite moves.

bmp_objects	rmb sizeof_bmp
bmp_objects_end

objects
doors		rmb max_doors*sizeof_obj
doors_end
pickups
keys		rmb max_keys*sizeof_obj
keys_end
items		rmb max_items*sizeof_obj
items_end
pickups_end
trapdoors	rmb max_trapdoors*sizeof_obj
trapdoors_end
objects_end
shots		rmb 12*sizeof_obj
shots_end
all_objects_end

; Modified by PAL/NTSC switching

vdg_mode	rmb 1
logo		rmb 2
play_screen_dz	rmb 2
text_screen_dz	rmb 2

; Song playback

p_timer		rmb 1
c1_a_timer	rmb 1
c1_f_timer	rmb 1
c2_a_timer	rmb 1
c2_f_timer	rmb 1
c3_a_timer	rmb 1
c3_f_timer	rmb 1

psg_c1att	rmb 1
psg_c2att	rmb 1
psg_c3att	rmb 1

song_ptr	rmb 2
c1_a_next	rmb 2
c1_f_next	rmb 2
c2_a_next	rmb 2
c2_f_next	rmb 2
c3_a_next	rmb 2
c3_f_next	rmb 2

tbl_tiles_a_start

		rmb 2	; tile_door1_v_a
		rmb 2	; tile_door1_h_a
		rmb 2	; tile_door_v_a
		rmb 2	; tile_door_h_a

tbl_tiles_a

tile_floor_a	rmb 2	; tile_floor_a
		rmb 2	; tile_exit_a
		rmb 2	; tile_trapdoor_a
		rmb 2	; tile_food_a
		rmb 2	; tile_tport_a
		rmb 2	; tile_armour_a
		rmb 2	; tile_potion_a
		rmb 2	; tile_weapon_a
		rmb 2	; tile_cross_a
		rmb 2	; tile_speed_a
mod_drainer_a	rmb 2	; tile_drainer0_a
mod_money_a	rmb 2	; tile_money0_a
mod_power_a	rmb 2	; tile_power0_a
tile_drainer1_a	rmb 2	; XXX alignment
		rmb 2	; tile_key_a
		rmb 2	; tile_key1_a

		rmb 2	; tile_p1_up0_a
		rmb 2	; tile_p1_up1_a
		rmb 2	; tile_p1_down0_a
		rmb 2	; tile_p1_down1_a
		rmb 2	; tile_p1_left0_a
		rmb 2	; tile_p1_left1_a
		rmb 2	; tile_p1_right0_a
		rmb 2	; tile_p1_right1_a
		rmb 2	; tile_arrow_up_a
		rmb 2	; tile_arrow_down_a
		rmb 2	; tile_arrow_left_a
		rmb 2	; tile_arrow_right_a
		rmb 2	; tile_p1_exit4_a
		rmb 2	; tile_p1_exit5_a
		rmb 2	; tile_p1_exit6_a
		rmb 2	; tile_p1_exit7_a
		rmb 2	; tile_exit8_a
		rmb 2	; tile_exit9_a
		rmb 2	; tile_exit10_a
		rmb 2	; tile_exit11_a

		rmb 2	; tile_p2_up0_a
		rmb 2	; tile_p2_up1_a
		rmb 2	; tile_p2_down0_a
		rmb 2	; tile_p2_down1_a
		rmb 2	; tile_p2_left0_a
		rmb 2	; tile_p2_left1_a
		rmb 2	; tile_p2_right0_a
		rmb 2	; tile_p2_right1_a
		rmb 2	; tile_fball_up_a
		rmb 2	; tile_fball_down_a
		rmb 2	; tile_fball_left_a
		rmb 2	; tile_fball_right_a
		rmb 2	; tile_p2_exit4_a
		rmb 2	; tile_p2_exit5_a
		rmb 2	; tile_p2_exit6_a
		rmb 2	; tile_p2_exit7_a
		rmb 2	; tile_exit8_a
		rmb 2	; tile_exit9_a
		rmb 2	; tile_exit10_a
		rmb 2	; tile_exit11_a

		rmb 2	; tile_p3_up0_a
		rmb 2	; tile_p3_up1_a
		rmb 2	; tile_p3_down0_a
		rmb 2	; tile_p3_down1_a
		rmb 2	; tile_p3_left0_a
		rmb 2	; tile_p3_left1_a
		rmb 2	; tile_p3_right0_a
		rmb 2	; tile_p3_right1_a
		rmb 2	; tile_axe_up_a
		rmb 2	; tile_axe_down_a
		rmb 2	; tile_axe_left_a
		rmb 2	; tile_axe_right_a
		rmb 2	; tile_p3_exit4_a
		rmb 2	; tile_p3_exit5_a
		rmb 2	; tile_p3_exit6_a
		rmb 2	; tile_p3_exit7_a
		rmb 2	; tile_exit8_a
		rmb 2	; tile_exit9_a
		rmb 2	; tile_exit10_a
		rmb 2	; tile_exit11_a

		rmb 2	; tile_p4_up0_a
		rmb 2	; tile_p4_up1_a
		rmb 2	; tile_p4_down0_a
		rmb 2	; tile_p4_down1_a
		rmb 2	; tile_p4_left0_a
		rmb 2	; tile_p4_left1_a
		rmb 2	; tile_p4_right0_a
		rmb 2	; tile_p4_right1_a
		rmb 2	; tile_sword_up_a
		rmb 2	; tile_sword_down_a
		rmb 2	; tile_sword_left_a
		rmb 2	; tile_sword_right_a
		rmb 2	; tile_p4_exit4_a
		rmb 2	; tile_p4_exit5_a
		rmb 2	; tile_p4_exit6_a
		rmb 2	; tile_p4_exit7_a
		rmb 2	; tile_exit8_a
		rmb 2	; tile_exit9_a
		rmb 2	; tile_exit10_a
		rmb 2	; tile_exit11_a

		rmb 2	; tile_monster_up0_a
		rmb 2	; tile_monster_up1_a
		rmb 2	; tile_monster_down0_a
		rmb 2	; tile_monster_down1_a
		rmb 2	; tile_monster_left0_a
		rmb 2	; tile_monster_left1_a
		rmb 2	; tile_monster_right0_a
		rmb 2	; tile_monster_right1_a

		rmb 2	; tile_bones0_a
		rmb 2	; tile_bones1_a
		rmb 2	; tile_bones2_a
		rmb 2	; tile_bones3_a
		rmb 2	; tile_bones4_a
		rmb 2	; tile_bones5_a
		rmb 2	; tile_bones6_a
		rmb 2	; tile_bones7_a

tbl_tiles_a_end

tile_drainer0_a		rmb 2
tile_money0_a		rmb 2
tile_power0_a		rmb 2
tile_money1_a		rmb 2
tile_power1_a		rmb 2

tbl_tiles_b_start
tbl_tiles_b	equ *+tbl_tiles_a-tbl_tiles_a_start
tile_floor_b	equ *+tile_floor_a-tbl_tiles_a_start
mod_drainer_b	equ *+mod_drainer_a-tbl_tiles_a_start
mod_money_b	equ *+mod_money_a-tbl_tiles_a_start
mod_power_b	equ *+mod_power_a-tbl_tiles_a_start
tile_drainer1_b	equ *+tile_drainer1_a-tbl_tiles_a_start
		rmb tbl_tiles_a_end-tbl_tiles_a_start
tbl_tiles_b_end

tile_drainer0_b		rmb 2
tile_money0_b		rmb 2
tile_power0_b		rmb 2
tile_money1_b		rmb 2
tile_power1_b		rmb 2

tile_yes_b		rmb 2
tile_no_b		rmb 2

tile_chalice_nw0_a	rmb 2
tile_chalice_nw1_a	rmb 2

tile_stone01_a		rmb 2

tbl_lnum		rmb 10*2
snum_tiles_start	rmb 2
