; Dunjunz - Games Master Edition
; Copyright (c) 2017-2020 Ciaran Anscomb
;
; See COPYING file for redistribution conditions.

; ========================================================================

; There are 4 x 6 rooms in each map.  Each room has 8 x 8 positions.  Room
; IDs are organised like this:

; -$80  -$50  -$20  $10
; -$78  -$48  -$18  $18
; -$70  -$40  -$10  $20
; -$68  -$38  -$08  $28
; -$60  -$30   $00  $30
; -$58  -$28   $08  $38

; Object positions (y,x) are not strictly cartesian.  The 'y' coordinate
; is a room ID plus vertical offset within the room.  The 'x' coordinate
; is just horizontal offset into the room.  Thus, 'y' ranges from -$80 to
; $3f, and 'x' ranges from 0 to 7.
;
; Two simple bitmaps are maintained.  'bmp_wall' tracks where walls and
; doors are (anything that blocks player movement).  Doors are cleared
; within the bitmap when opened.  'bmp_objects' indicates where objects
; are, consulted both for pickups and undrawing.
;
; In each bitmap, one byte represents all positions for one 'y'
; coordinate.  Most significant bit maps to leftmost position ('x' == 0).

; ========================================================================

; Definitions

		include	"dragonhw.s"
		include	"games-master.s"
		include "objects.s"
		include "defs.s"
		include "macros.s"

		include "bank0a.exp"
		include "bank1.exp"
		include "bank2.exp"
		include "bank3.exp"

; ========================================================================

		org BSS_end

; Explicitly order the sections by listing them here.

CODE_start_	CODE
DATA_start_	DATA
		CODE

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Programmable Sound Generator

psg_latch_all
		ldx psg_latch_c1
		beq 10F
		jsr psg_submit
10		ldx psg_latch_c2
		beq 20F
		bsr psg_submit
20		ldx psg_latch_c3
		beq 30F
		bsr psg_submit
30		clra
		clrb
		std psg_latch_c1
		std psg_latch_c2
		std psg_latch_c3
		; fall through to psg_playfrag

; This sends data to the SN76489.

psg_playfrag
		pshs u

		lda #200
10		deca
		bne 10B

		ldu #psg_c1
		bsr psg_playchan
		ldu #psg_c2
		bsr psg_playchan
		ldu #psg_c3
		bsr psg_playchan
		puls u,pc

psg_playchan
		; handle tone data
		dec psg_counter,u
		bne 40F
		ldx psg_ptr,u
		lda ,x+
		beq 20F
10		ldb ,x+
		stb reg_gmc_snd
		deca
		bne 10B
		lda ,x+
		bne 30F
20		ldx ,x
		inca
30		sta psg_counter,u
		stx psg_ptr,u
		; handle attenuation envelope
40		dec psg_ecounter,u
		bne 50F
		ldx psg_eptr,u
		ldd ,x++
		sta psg_arate,u
		bne 45F
		com psg_atten,u
45		stb psg_ecounter,u
		beq 50F
		stx psg_eptr,u
50		lda psg_arate,u
		bmi 60F
		adda psg_atten,u
		cmpa psg_atten,u
		bhs 70F
		lda #$ff
		bra 70F
60		adda psg_atten,u
		cmpa psg_atten,u
		bls 70F
		clra
70		sta psg_atten,u
		lsra
		lsra
		lsra
		lsra
		ora psg_areg,u
		sta reg_gmc_snd
		rts
		

		; hack to determine channel without changing loads of
		; calls to psg_submit...  do this properly!
psg_submit
		pshs y,u
		ldy ,x++
		lda ,x+
		bita #$40
		bne 20F
		bita #$20
		bne 10F
		ldu #psg_c1
		bra 30F
10		ldu #psg_c2
		bra 30F
20		ldu #psg_c3
30		tfr a,b
		anda #$f0
		sta psg_areg,u
		lslb
		lslb
		lslb
		lslb
		stb psg_atten,u
		sty psg_eptr,u
		stx psg_ptr,u
		ldb #$01
		stb psg_counter,u
		stb psg_ecounter,u
		puls y,u,pc

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Main game loop

		; tile set 0
anim_tiles_0	ldx tile_drainer0_a
		stx mod_drainer_a
		ldx tile_drainer0_b
		stx mod_drainer_b
		ldx tile_money0_a
		stx mod_money_a
		ldx tile_money0_b
		stx mod_money_b
		ldx tile_power0_a
		ldu tile_power0_b
		ldy tile_chalice_nw0_a
		bra 10F

		; tile set 1
anim_tiles_1	ldx tile_drainer1_a
		stx mod_drainer_a
		ldx tile_drainer1_b
		stx mod_drainer_b
		ldx tile_money1_a
		stx mod_money_a
		ldx tile_money1_b
		stx mod_money_b
		ldx tile_power1_a
		ldu tile_power1_b
		ldy tile_chalice_nw1_a
10		stx mod_power_a
		stu mod_power_b
		sty mod_chalice
mod_draw_chalice_0
                jsr update_chalice
		bra anim_done

; Two-frame animations toggle every 8 frames.  This is called from
; 'mainloop' below, and also during the end animation so that any drainers
; still visible continue to animate.

animate		inc anim_count
		lda anim_count
		anda #15
		beq anim_tiles_1
		anda #7
		beq anim_tiles_0
		; Redraw select range of object types every call.
anim_done	ldu #pickups
10		lda obj_tile,u
		cmpa #always_draw_tile
		blo 20F
		jsr obj_draw
20		leau sizeof_obj,u
		cmpu #all_objects_end
		blo 10B
		rts

mainloop

		jsr psg_playfrag
		bsr animate
		jsr psg_latch_all

		jsr process_monsters
		jsr plr_scan_controls
		jsr psg_playfrag
		jsr process_players
		jsr process_shots

		; really have to slow things down...
		jsr psg_playfrag
		ldx #$0200
!		leax -1,x
		bne <

		ldd #$fb40
		sta reg_pia0_pdrb
		bitb reg_pia0_pdra	; BREAK?
		bne 10F

		lda #$fe
		sta reg_pia0_pdrb
		bitb reg_pia0_pdra	; ENTER?
		beq 90F
		bra pause

10

		lda finished
		ora dead
		cmpa #$0f
		bne mainloop

		dec finish_delay
		bne mainloop

		lda dead
		cmpa playing
		bne finished_level

		lda #$ff
		sta level	; no cheating once all dead!
		ldy #death_screen	; dzip start
game_over_screen
		ldb colour_fg
		stb text_colour
		jsr unpack_text_screen
		lda #3
		sta reg_gmc_bank
		jsr write_screen

		; a few seconds pause for reflection
		ldx #$2000
20		deca
		bne 20B
		leax -1,x
		bne 20B
90		jmp game_reset

pause
		lda #3
		sta reg_gmc_bank
		jsr play_song
		lda #2
		sta reg_gmc_bank
		bra 10B

finished_level
		lda level
		cmpa #24
		beq end_animation
		jmp level_setup

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Ending animation.  A bit naff to be honest, just spins the dying monster
; animation around the chalice for a bit (like the original) then just
; ends.  Would be nice to squeeze in something a bit more impressive.

eanim_add
		lda obj_y,y
		cmpa #$80
		bne 99F
		pulu a,b
		std obj_y,y
		stb spr_cx,y
		jsr yx_to_woff
		stx obj_woff,y
		stx spr_owoff,y
		clr spr_state,y
		leay sizeof_mon,y
99		rts

mon_step4	bsr mon_step1
		bsr mon_step1
		bsr mon_step1
mon_step1	pshs u
		jsr mon_run_state
		puls u,pc

eanim_step	pshs u
		ldu #monsters
10		lda obj_y,u
		deca
		bvs 20F
		lda spr_state,u
		ldx #jtbl_eanim_state
		jsr [a,x]
20		leau sizeof_mon,u
		cmpu #monsters_end
		blo 10B
		puls u,pc

end_animation
		; first, destroy all monsters
		ldu #monsters
		ldb #state_mon_die
10		lda obj_y,u
		deca
		bvs 20F
		stb spr_state,u
20		leau sizeof_mon,u
		cmpu #monsters_end
		blo 10B
		; step monsters through to death
		bsr mon_step4

		ldd #1200
		std mod_edelay

		lda #2
		sta mod_eanim_rate

		clr ,-s

30		ldy #monsters
		ldu #tbl_eanim

40		bsr eanim_add
		pshs y,u

		lda 4,s
		anda #127
		bne 42F
		ldx #snd_tport
		stx psg_latch_c1
42		anda #15
		bne 44F
		ldx #snd_psh
		stx psg_latch_c3
44		inc 4,s
		jsr psg_latch_all

		jsr animate
		bsr eanim_step

		ldb #2
		pshs b
49		jsr magic_playfrag
mod_edelay	equ *+1
		ldx #1200
50		leax -1,x
		bne 50B
		dec ,s
		bne 49B
		leas 1,s

		ldx mod_edelay
		leax -2,x
		beq 90F
		stx mod_edelay
		cmpx #900
		beq 60F
		cmpa #600
		bne 62F
60		lda mod_eanim_rate
		adda #2
		sta mod_eanim_rate
62

		puls y,u
		cmpu #tbl_eanim_end
		blo 40B
		bra 30B

90		leas 5,s

		; finish off
		ldb #15
		stb tmp2+1
92		jsr animate
		jsr eanim_step
		jsr psg_playfrag
		dec tmp2+1
		bne 92B

		jmp level_setup

eanim_start
		lda #bones0
		bra 10F
eanim_cont
		lda obj_tile,u
		inca
		cmpa #bones2
		bls 10F
		clra
10		sta obj_tile,u
		ldx obj_woff,u
		ldb obj_x,u
		jsr obj_draw_vx
		lda obj_tile,u
		bne 20F
		lda #$80
		sta obj_y,u
20		lda spr_state,u
mod_eanim_rate	equ *+1
		adda #2
		sta spr_state,u
		rts

		DATA
tbl_eanim
		fcb $16,4
		fcb $16,3
		fcb $16,2
		fcb $16,1
		fcb $15,1
		fcb $14,1
		fcb $13,1
		fcb $13,2
		fcb $13,3
		fcb $13,4
		fcb $14,4
		fcb $15,4
tbl_eanim_end

jtbl_eanim_state
		STATE_JTBL
		STATE_ID "state_eanim_start"
		fdb eanim_start
		fdb spr_delay
		fdb spr_delay
		fdb spr_delay
		fdb eanim_cont
		fdb spr_delay
		fdb spr_delay
		fdb spr_delay
		fdb eanim_cont
		fdb spr_delay
		fdb spr_delay
		fdb spr_delay
		fdb eanim_cont
		fdb spr_nop
		fdb spr_nop
		fdb spr_nop
		CODE


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; First state while changing room.  Initialises player window with map
; outline: stone & floor.  Next state will draw all objects.

; Faster drawing, as can make more assumptions when doing every tile.

; Entry:
;     X -> screen
;     U -> player

DRAW_ROW_A_FAST	macro
		pulu d
		std \1,x
		endm

DRAW_LAST_ROW_A_FAST	macro
		ldd ,u
		std \1,x
		endm

DRAW_ROW_B_FAST	macro
		lda \1,x
		clrb
		addd ,u++
		std \1,x
		endm

DRAW_LAST_ROW_B_FAST	macro
		lda \1,x
		clrb
		addd ,u
		std \1,x
		endm

plr_draw_room
		ldb spr_next_y,u
		andb #$f8
		ldx #bmp_wall+128
		leay b,x
		ldx plr_win,u
		lda #win_h
		sta tmp1
		pshs u
10		lda ,y+
		sta tmp0
		lda #win_w/2
		sta tmp1+1

		; even tiles
20		ldu tile_floor_a
		lsl tmp0
		bcc 30F
mod_stone_a	equ *+1
		ldu #$0000
30		DRAW_ROW_A_FAST -128
		DRAW_ROW_A_FAST -96
		DRAW_ROW_A_FAST -64
		DRAW_ROW_A_FAST -32
		DRAW_ROW_A_FAST 0
		DRAW_ROW_A_FAST 32
		DRAW_ROW_A_FAST 64
		DRAW_ROW_A_FAST 96
		DRAW_LAST_ROW_A_FAST 128

		; odd tiles
		leax 1,x
		ldu tile_floor_b
		lsl tmp0
		bcc 40F
mod_stone_b	equ *+1
		ldu #$0000
40		DRAW_ROW_B_FAST -128
		DRAW_ROW_B_FAST -96
		DRAW_ROW_B_FAST -64
		DRAW_ROW_B_FAST -32
		DRAW_ROW_B_FAST 0
		DRAW_ROW_B_FAST 32
		DRAW_ROW_B_FAST 64
		DRAW_ROW_B_FAST 96
		DRAW_LAST_ROW_B_FAST 128

		leax 2,x
		dec tmp1+1
		lbne 20B

		leax (9*fb_w)-12,x
		dec tmp1
		lbne 10B
		puls u
		jmp spr_inc_state

DRAW_ROW_A	macro
		clra
		ldb (\1+1),x
		andb #$0f
		addd ,u++
		std \1,x
		endm

DRAW_LAST_ROW_A	macro
		clra
		ldb (\1+1),x
		andb #$0f
		addd ,u
		std \1,x
		endm

DRAW_ROW_B	macro
		lda \1,x
		clrb
		anda #$f0
		addd ,u++
		std \1,x
		endm

DRAW_LAST_ROW_B	macro
		lda \1,x
		clrb
		anda #$f0
		addd ,u
		std \1,x
		endm

single_tile_a	ldu #tbl_tiles_a
		leau a,u
		ldu a,u
single_tile_x_a
		DRAW_ROW_A -128
		DRAW_ROW_A -96
		DRAW_ROW_A -64
		DRAW_ROW_A -32
		DRAW_ROW_A 0
		DRAW_ROW_A 32
		DRAW_ROW_A 64
		DRAW_ROW_A 96
		DRAW_LAST_ROW_A 128
		rts

draw_object_once
		ldb obj_x,y
		lsrb
		bcc single_tile_a
		; fall through to single_tile_b

single_tile_b	ldu #tbl_tiles_b
		leau a,u
		ldu a,u
single_tile_x_b
		DRAW_ROW_B -128
		DRAW_ROW_B -96
		DRAW_ROW_B -64
		DRAW_ROW_B -32
		DRAW_ROW_B 0
		DRAW_ROW_B 32
		DRAW_ROW_B 64
		DRAW_ROW_B 96
		DRAW_LAST_ROW_B 128
99		rts

update_chalice
		ldu #player1
		bsr plr_update_chalice
		ldu #player2
		bsr plr_update_chalice
		ldu #player3
		bsr plr_update_chalice
		ldu #player4
		; fall through to plr_update_chalice

plr_update_chalice
		ldb plr_yroom,u
plr_update_chalice_b
		andb #$f8
		cmpb #$10
		bne 99B
10		ldx plr_win,u
		leax 4*9*fb_w+3,x
mod_chalice	equ *+1
		ldu #tile_chalice_nw0_a
		jsr single_tile_x_a
		leau 2,u
		leax 1,x
		jsr single_tile_x_b
		leau 2,u
		leax 9*fb_w-1,x
		jsr single_tile_x_a
		leau 2,u
		leax 1,x
		jmp single_tile_x_b

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		setdp 0			; assumed
game_reset	nop

		orcc #$50
		lds #$0100

		lda #$ff
		tfr a,dp
		setdp $ff

		; A == $ff from above
		clr reg_pia0_cra	; $00 - DDR
		clr reg_pia0_crb	; $00 - DDR
		clr reg_pia1_cra	; $00 - DDR
		clr reg_pia1_crb	; $00 - DDR
		ldb #$35		; $35 - PDR, FS hi->lo
		std reg_pia0_ddrb	; $ff - all outputs
		clra			; $00 - all inputs
		decb			; $34 - PDR, no HS
		std reg_pia0_ddra
		lda #$fc		; $fc - only DAC bits as outputs
		std reg_pia1_ddra	; $34 - PDR, no printer FIRQ
PATCH_p1ddrb	equ *+1
		ldd #$fc3c		; $fc - VDG, ROMSEL as outputs
		std reg_pia1_ddrb	; $3c - PDR, no CART FIRQ

		; SAM MPU rate = slow
		sta reg_sam_r1c
		sta reg_sam_r0c

		; SAM display offset = $0200
		std reg_sam_f0s		; and f1c
		sta reg_sam_f2c

		; SAM VDG mode = G6R,G6C
		sta reg_sam_v0c
		sta reg_sam_v1s
		sta reg_sam_v2s

		jsr set_cc3_palette

		; NEVER sit in map type 1
		sta reg_sam_tyc

		lda #gamedp
		tfr a,dp
		setdp gamedp

		lda level
		inca
		beq 10F
		ldd #$fd40
		sta reg_pia0_pdrb
		bitb reg_pia0_pdra	; CLEAR?
		lbeq cheat
10

show_title_screen

		lda #3
		sta reg_gmc_bank
		lda #$e2
		sta reg_pia1_pdrb       ; VDG mode CG6, CSS0
		ldx #fb0
		ldd #$aaaa
10		std ,x++
		cmpx #fb0_end
		blo 10B
		ldu #fb0+10*fb_w
		ldd #fb0+10*fb_w+151*fb_w
		std dzip_end
		ldx #title_bin_dz
		jsr ldunzip
		clr text_colour
		ldd #171*256+41
		ldy #press_key
		jsr write_at
		jsr reset_song
		jsr play_song

show_select_screen

		; three times around the select/credits screens before
		; returing to title
		lda #3
		sta tmp2

redraw_select_screen

		lda #$e2
		sta reg_pia1_pdrb	; VDG mode CG6, CSS0

		jsr unpack_text_screen

		lda #3
		sta reg_gmc_bank
		ldb colour_fg
		stb text_colour
		ldu #textlist_select_screen
		jsr write_textlist
		lda colour_hi
		sta text_colour
		jsr write_charlist

		jsr update_space
		jsr update_p1
		jsr update_p2
		jsr update_p3
		jsr update_p4
		jsr update_vmode

		ldd #$0a0a
		std player1+spr_tmp0
		std player2+spr_tmp0
		std player3+spr_tmp0
		std player4+spr_tmp0

select_loop
		jsr wait_keyheld

		ldd #$0600	; about 30s delay
		std tmp0	; until next screen

select_anim_loop

		; Selected players are animated, showing them moving in
		; one direction for a while, then moving in another.

		; spr_tmp0,u is the inter-frame animation counter
		; spr_tmp0+1,u counts between updates to facing direction

		lda #2
		sta reg_gmc_bank

		ldx #select_anim
		ldu #players
10		lda plr_bit,u
		bita playing
		beq 90F
		dec spr_tmp0,u
		bne 90F

		lda #10
		sta spr_tmp0,u
		lda spr_tilebase,u
		eora #1
		sta spr_tilebase,u

		dec spr_tmp0+1,u
		bne 90F
		lda #10
		sta spr_tmp0+1,u
		lda spr_facing,u
		lsra
		lda a,x
		sta spr_facing,u

90		leau sizeof_plr,u
		cmpu #players_end
		blo 10B

		; delay ~1/50s
		ldd #2400
10		subd #1
		bne 10B

		ldu #players
		ldx #fb0+53*fb_w+4+128
10		lda spr_tilebase,u
		adda spr_facing,u
		pshs x,u
		jsr single_tile_b
		puls x,u
		leax 14*fb_w,x
		leau sizeof_plr,u
		cmpu #players_end
		blo 10B

		lda #3
		sta reg_gmc_bank

		ldx #tbl_kdsp_menu
		jsr test_keypress_jump

		dec tmp0+1
		bne select_anim_loop
		dec tmp0
		bne select_anim_loop
		; fall through to show_credits

		; - - - -

		; Credits screen 1
show_credits	ldb colour_fg
		stb text_colour
		jsr unpack_text_screen
		lda #3
		sta reg_gmc_bank
		ldy #credits_screen
		jsr write_screen

		bsr credits_wait
		lbne show_select_screen

		; - - - -

		; Credits screen 2
		inc mod_line_height
		jsr unpack_text_screen
		lda #3
		sta reg_gmc_bank
		ldy #credits_screen2
		jsr write_screen
		dec mod_line_height

		bsr credits_wait
		lbne show_select_screen

		; - - - -

		; Return either to select screen or title screen.
		dec tmp2
		lbne redraw_select_screen
		jmp show_title_screen

		; - - - -

; Wait around 20s per credits screen
credits_wait	ldx #1000
10		ldd #2540
20		subd #1
		bne 20B
		leax -1,x
		beq 99F
		jsr test_keypressed
		beq 10B
99		rts

toggle_p1	lda #$01
		bsr toggle_player
		bsr update_p1
		jmp select_loop
toggle_p2	lda #$02
		bsr toggle_player
		bsr update_p2
		jmp select_loop
toggle_p3	lda #$04
		bsr toggle_player
		bsr update_p3
		jmp select_loop
toggle_p4	lda #$08
		bsr toggle_player
		bsr update_p4
		jmp select_loop
toggle_player	eora playing
		sta playing
		bsr update_space
		rts

update_p1	lda #$01
		ldx #fb0+53*fb_w+16+128
		bra update_player
update_p2	lda #$02
		ldx #fb0+67*fb_w+16+128
		bra update_player
update_p3	lda #$04
		ldx #fb0+81*fb_w+16+128
		bra update_player
update_p4	lda #$08
		ldx #fb0+95*fb_w+16+128
update_player	ldb #2
		stb reg_gmc_bank
		anda playing
		bne draw_yes
draw_no		ldu tile_no_b
		bra 10F
draw_yes	ldu tile_yes_b
10		jmp single_tile_x_b

toggle_skill	bsr undraw_skill
		com difficulty
		bsr update_skill
		jmp select_loop
toggle_vmode	jsr advance_video_mode
		jmp show_select_screen

; Changing video mode means redrawing screen, no need for "undraw".

update_vmode	lda #3
		sta reg_gmc_bank
		lda colour_fg
		sta text_colour
		lda video_mode
		lsla
		ldx #tbl_text_vmode
		ldy a,x
		ldd #132*256+62
		jsr write_at
		; only called in one place, and update_skill required
		; immediately after, so...
		; fall through to update_skill

update_skill	lda colour_fg
		bra 10F
undraw_skill	lda colour_bg
10		sta text_colour
		ldd #118*256+62
		tst difficulty
		beq 20F
		ldy #text_hard
		bra 30F
20		ldy #text_easy
30		jmp write_at

; "<Space> to play" is only highlighed when a player is selected.
update_space	lda playing
		bne 10F
		lda colour_fg
		bra 20F
10		lda colour_hi
20		sta text_colour
		ldd #160*256+32
		ldy #text_space
		jmp write_at

		DATA
; Select screen character animation: next "facing" based on current so
; that animation is in order: up, right, down, left.
select_anim	fcb 6,4,0,2

		include "text-ram.s"

		export tbl_text_vmode
tbl_text_vmode	fdb text_pal
		fdb text_ntsc0
		fdb text_ntsc1

tbl_kdsp_menu	KEYDSP_ENTRY 7,5,start_game	; SPACE
		KEYDSP_ENTRY 1,0,toggle_p1	; '1'
		KEYDSP_ENTRY 2,0,toggle_p2	; '2'
		KEYDSP_ENTRY 3,0,toggle_p3	; '3'
		KEYDSP_ENTRY 4,0,toggle_p4	; '4'
		KEYDSP_ENTRY 4,2,toggle_skill	; 'D'
		KEYDSP_ENTRY 6,4,toggle_vmode	; 'V'
		KEYDSP_ENTRY 2,6,game_reset	; BREAK
		fdb 0
		CODE

		; - - - - - - - - - - - - - - - - - - - - - - -

		DATA
easy_patchset	PATCH mod_skill_door
		fcb $02		; alternate door graphic
		ENDPATCH
		PATCH mod_skill_key
		fcb $4c		; alternate key graphic
		ENDPATCH
		PATCH mod_drainer_hp
		fcb $09		; lower drainer strength
		ENDPATCH
		PATCH mod_monster_hp
		fcb $04		; half initial hp
		ENDPATCH
		PATCH mod_mon_hit_player
		fcb $92		; 8-armour damage -> 100-8(BCD)
		ENDPATCH
		PATCH mod_drainer_dmg
		fcb $90		; 10 damage -> 100-10(BCD)
		ENDPATCH
		ENDPATCHSET

hard_patchset	PATCH mod_skill_door
		fcb $00		; don't alternate door graphic
		ENDPATCH
		PATCH mod_skill_key
		fcb $12		; don't alternate key graphic
		ENDPATCH
		PATCH mod_drainer_hp
		fcb $0c		; full drainer strength
		ENDPATCH
		PATCH mod_monster_hp
		fcb $08		; full initial hp
		ENDPATCH
		PATCH mod_mon_hit_player
		fcb $90		; 10-armour damage -> 100-10(BCD)
		ENDPATCH
		PATCH mod_drainer_dmg
		fcb $80		; 20 damage -> 100-20(BCD)
		ENDPATCH
		ENDPATCHSET
		CODE

start_game
		lda #1
		sta reg_gmc_bank

		; must select some players
		lda playing
		lbeq select_loop

		ldu #hard_patchset
		lda difficulty
		bne 10F
		ldu #easy_patchset
10		jsr apply_patchset

		; flag all players as "dead" so they get reset
		lda #$ff
		sta dead

	if DEBUG_LEVEL25

		bra 30F
cheat		lda #23
30		sta level

	else

		; level -1: will be incremented before use
		sta level
cheat

	endif

		; - - - - - - - - - - - - - - - - - - - - - - -

		; Set up next level.

level_setup

		lda #2
		sta reg_gmc_bank

		ldx #snd_reset_c1
		stx psg_latch_c1
		ldx #snd_reset_c2
		stx psg_latch_c2
		ldx #snd_reset_c3
		stx psg_latch_c3
		jsr psg_latch_all

		lda #1
		sta reg_gmc_bank

		lda playing
		eora #$0f
		sta finished
		clr exiting

		lda #30
		sta finish_delay

		; Clear monsters
		ldx #monsters
		ldd #$8000|monster_tilebase
10		clr spr_hp,x
		sta obj_y,x
		stb spr_tilebase,x
		leax sizeof_mon,x
		cmpx #monsters_end
		blo 10B

		; Next level.  Detect when game is complete (finished
		; level 25).  Modify various routines to only call
		; chalice-based routines on level 25.  Decode level data.
		ldb #$8c	; cmpx
		inc level
		lda level
		cmpa #24
		beq 10F
		blo 20F
		ldy #end_screen
		jmp game_over_screen
10		ldb #$bd	; jsr
20		stb mod_draw_chalice_0
		stb mod_draw_chalice_1
		stb mod_check_chalice

		ldu #levels
		; level size is 378 but we can't multiply by that...
		ldb #189
		lsla
		mul
		leau d,u
		; copy initial wall bitmap
		ldx #bmp_wall
25		ldd ,u++
		std ,x++
		cmpx #bmp_wall_end
		blo 25B
		; clear object bitmap
30		clr ,x+
		cmpx #bmp_objects_end
		blo 30B

		; Select stone graphic based on level
		lda level
		ldb #18
		mul
		addd tile_stone01_a
		std mod_stone_a
		addd #25*18
		std mod_stone_b

		; Populate object table

		; source: u left pointing to start of level object list
		ldy #objects		; destination: object table

		; Doors.  Source is (y,x) pairs.  x is shifted left one
		; bit, with the bottom bit then used to indicate
		; horizontal or vertical door.
		clr tmp0+1
		lda #max_doors
		sta tmp0
10		pulu d
		cmpd #$8080
		bne 20F
		; invalid
		tfr d,x
		clra
		bra 40F
		; valid
20		lsrb		; CC.C set => vertical door
		tfr d,x
		lda #door_h
		bcc 30F
		lda #door_v
30		com tmp0+1
		bne 40F
mod_skill_door	equ *+1
		suba #$00	; change to suba #2 to alternate colours
40		bsr place_object
		jsr clr_pickup_x	; not a pickup
		dec tmp0
		bne 10B

		; Keys.  Source is (y,x) of key, and order is important:
		; the nth key here opens the nth door above.
		clr tmp0+1
		lda #max_keys
		sta tmp0
10		pulu x
		clra
		cmpx #$8080
		beq 20F
		lda #key
		com tmp0+1
		bne 20F
mod_skill_key	nop		; change to inca to alternate colours
20		bsr place_object
		; fetch door location from earlier in map data
		ldd -(max_doors*2)-2,u
		lsrb
		std obj_key_opens-sizeof_obj,y
30		dec tmp0
		bne 10B

		; Items (pickups, drainers).  Source is (type,y,x).
		lda #max_items
		sta tmp0
10		pulu a,x
		; ignored for non-drainers
mod_drainer_hp	equ *+1
		ldb #12
		stb obj_drainer_hp,y
		bsr place_object
		dec tmp0
		bne 10B

		; Trapdoors.  Source is (y,x).
		lda #max_trapdoors
		sta tmp0
10		pulu x
		lda #trapdoor
		bsr place_object
		dec tmp0
		bne 10B

		; Clear shots.
		clra
10		bsr place_object
		cmpy #shots_end
		blo 10B

		bra finish_setup

; Set object type & location, mark pickup bitmap if valid (type != 0).

; Entry:
; 	A = object type
; 	X = y,x
; 	Y -> object
place_object
		pshs x
		stx obj_y,y
		sta obj_tile,y
		beq 90F
		tfr x,d
		jsr set_pickup
		jsr yx_to_woff
		stx obj_woff,y
90		leay sizeof_obj,y
		puls x,pc

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		export first_init_players

		; JMPed to from ROM as last part of initialisation
first_init_players
		bsr init_players
		jmp game_reset

; Set up players.  If player is dead, restore default stats and clear
; score.  Always restore to 99 health.

init_players

		lda #1
		sta reg_gmc_bank
		ldy #player1
		ldu #tbl_p_dfl
		lda #1

10
		sta plr_bit,y

		pulu a,b,x
		sta spr_tilebase,y
		stb spr_facing,y	; initial direction
		stx spr_next_y,y	; incl. spr_next_x
		pulu a,b,x
		std plr_swin,y		; stats window address
		stx plr_win,y		; player window address

		ldd #$8080
		std obj_y,y
		std plr_key_opens,y	; no key
		clr plr_speed,y		; no speedyboots
		ldd #$9909		; init health
		sta spr_hp,y		; init health = 99
		stb plr_yroom,y		; invalidate yroom
		clr plr_nshots,y	; not shooting
		; initial state for all players is to draw room
		ldb #state_plr_draw_room
		stb spr_state,y

		pulu a,x
		lsr dead		; player dead?
		bcc 20F
		; reset player stats if dead
		sta plr_armour,y
		stx plr_power,y		; incl. ammo
		clr plr_score2,y	; clear score
		clr plr_score0,y
20

		lda plr_bit,y
		leay sizeof_plr,y
		lsla
		cmpa #$08
		bls 10B

		rts

; Continues from level_setup.

finish_setup
		bsr init_players
		jsr unpack_play_screen

		clra
		clrb

		sta dead		; clear remaining bits
		sta dying		; and dying flags
		ldx #death_fifo		; initialise death fifo
		stx death_fifo_end

		lda #2
		sta reg_gmc_bank

		jmp mainloop

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Player input.  Scan keyboard & joysticks, set plr_control accordingly.

		setdp $ff	; only ever called with DP=$ff

		DATA
tbl_p1_keys	KEYTBL_ENTRY 3,2,2			; 'C' = magic
		KEYTBL_ENTRY 7,4,$80|facing_up		; 'W' = up
		KEYTBL_ENTRY 3,4,$80|facing_down	; 'S' = down
		KEYTBL_ENTRY 0,5,$80|facing_right	; 'X' = right
		KEYTBL_ENTRY 2,5,$80|facing_left	; 'Z' = left
		KEYTBL_ENTRY 1,4,1			; 'Q' = fire
		fdb 0
tbl_p2_keys	KEYTBL_ENTRY 2,2,2			; 'B' = magic
		KEYTBL_ENTRY 7,3,$80|facing_up		; 'O' = up
		KEYTBL_ENTRY 3,3,$80|facing_down	; 'K' = down
		KEYTBL_ENTRY 5,3,$80|facing_right	; 'M' = right
		KEYTBL_ENTRY 6,3,$80|facing_left	; 'N' = left
		KEYTBL_ENTRY 0,4,1			; 'P' = fire
		fdb 0
		CODE

; On exit, B always == 0

scan_keys
10		ldd ,x++
		beq 99F
		sta reg_pia0_pdrb
		lda ,x+
		andb reg_pia0_pdra
		bne 10B
		sta plr_control,u
99		rts

plr_scan_controls

		setdp gamedp		; called with usual game DP

		clra
		sta player1+plr_control
		sta player2+plr_control
		sta player3+plr_control
		sta player4+plr_control

		; keyboard & joystick scanning - will hit the PIAs a lot,
		; so set DP
		coma		; lda #$ff
		tfr a,dp
		setdp $ff

		; player 1 - table-driven scan
		ldu #player1
		ldx #tbl_p1_keys
		bsr scan_keys

		; player 2 - table-driven scan
		ldu #player2
		ldx #tbl_p2_keys
		bsr scan_keys

		decb			; ldb #$ff - scan_keys leaves b==0
		stb reg_pia0_pdrb	; polling firebuttons

		lda #$34
		sta reg_pia1_crb

		; player 3 - left joystick
		ldu #player3
		lda #$3d
		sta reg_pia0_crb	; left joystick
		lda reg_pia0_pdra
		bita #$02
		bne 10F
		lda #1			; fire
		sta plr_control,u
10		ldb p3_moved
		beq 30F
		bsr joy_yaxis
		bpl 20F
		bsr spr_move_control
		beq 50F
20		bsr joy_xaxis
		bra 50F
30		bsr joy_xaxis
		bpl 40F
		bsr spr_move_control
		beq 50F
40		bsr joy_yaxis
50

		; player 4 - right joystick
		ldu #player4
		lda #$35
		sta reg_pia0_crb	; right joystick
		lda reg_pia0_pdra
		bita #$01
		bne 10F
		lda #1			; fire
		sta plr_control,u
10		ldb p4_moved
		beq 30F
		bsr joy_yaxis
		bpl 20F
		bsr spr_move_control
		beq 50F
20		bsr joy_xaxis
		bra 50F
30		bsr joy_xaxis
		bpl 40F
		bsr spr_move_control
		beq 50F
40		bsr joy_yaxis
50

		ldd #$343f
		sta reg_pia0_cra
		stb reg_pia0_crb
		lda #$3c
		sta reg_pia1_crb

		lda #gamedp
		tfr a,dp

		rts

; joy_yaxis and joy_xaxis return with CC.N set if direction is detected.

joy_yaxis	lda #$3c
		sta reg_pia0_cra	; y axis
		ldd #$8040
		sta reg_pia1_pdra
		stb reg_pia1_pdra
		lda reg_pia0_pdra
		bmi 20F
		lda #$80|facing_up	; up
		bra 80F
20		ldd #$ffc0
		sta reg_pia1_pdra
		stb reg_pia1_pdra
		lda reg_pia0_pdra
		bpl 99F
		lda #$80|facing_down	; down
		bra 80F

; just about fits between the two axis checks to make all branches short
spr_move_control
		anda #$7f
		ldx obj_y,u
		pshs x
		ldx spr_next_y,u
		stx obj_y,u
		jsr spr_move_a
		pshs cc
		ldx 1,s
		stx obj_y,u
		puls cc,x,pc

joy_xaxis	lda #$34
		sta reg_pia0_cra	; x axis
		ldd #$8040
		sta reg_pia1_pdra
		stb reg_pia1_pdra
		lda reg_pia0_pdra
		bmi 20F
		lda #$80|facing_left	; left
		bra 80F
20		ldd #$ffc0
		sta reg_pia1_pdra
		stb reg_pia1_pdra
		lda reg_pia0_pdra
		bpl 99F
		lda #$80|facing_right	; right
80		sta plr_control,u
99		rts

		setdp gamedp

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Unpack dzipped screen and draw player stats to it.

plr_draw_hp	ldx plr_swin,u
		leax 13*fb_w+128,x
		lda spr_hp,u
		bra draw_lnum_pair

; Entry:
; 	U -> player
plr_draw_stats	pshs y,u
		lda #2
		sta reg_gmc_bank
		ldx plr_swin,u
		leax fb_w+128,x
		lda spr_tilebase,u
		adda #2
		jsr single_tile_a
		ldu 2,s
		jsr plr_draw_armour
		jsr plr_draw_power
		bsr plr_draw_ammo
		bsr plr_draw_hp
		bsr plr_draw_score
		puls y,u,pc

plr_add_score
		adda plr_score0,u
		daa
		sta plr_score0,u
		lda plr_score2,u
		adca #$00
		daa
		sta plr_score2,u
		; fall through to plr_draw_score

plr_draw_score	ldx plr_swin,u
		leax 23*fb_w+128,x
		lda plr_score2,u
		bsr draw_lnum_pair
		lda plr_score0,u
		; fall through to draw_lnum_pair

draw_lnum_pair
		pshs a
		lsra
		lsra
		lsra
		bsr 10F
		puls a
		lsla
10		anda #$1e
		pshs u
		ldu #tbl_lnum
		ldu a,u
		pulu d
		sta -128,x
		stb -96,x
		pulu d
		sta -64,x
		stb -32,x
		pulu d
		sta ,x+
		stb 31,x
		lda ,u
		sta 63,x
		puls u,pc

plr_draw_ammo	ldx plr_swin,u
		leax 15*fb_w+2,x
		lda plr_ammo,u
		sta tmp1
		suba #4
		sta tmp1+1
PATCH_draw_ammo0	equ *+1
		ldd #$55ba
		stb 1,x
		leax fb_w,x
PATCH_draw_ammo1	equ *+1
10		ldb #$7e
		std ,x
PATCH_draw_ammo2	equ *+1
		ldb #$ba
		stb fb_w+1,x
		leax 2*fb_w,x
		dec tmp1
		bne 10B
PATCH_draw_ammo3	equ *+1
		ldd #$aaaa
20		inc tmp1+1
		beq 99F
		std ,x
		std fb_w,x
		leax 2*fb_w,x
		bra 20B
99		rts

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Monster and player state machine handling.

; Monsters.  First, spawn a new monster if a slot is available, then for
; each extant monster, run its state machine.

process_monsters
		bsr new_monster
mon_run_state
		ldu #monsters
10		lda obj_y,u
		deca
		bvs 20F
		lda spr_state,u
		ldx #jtbl_monster_state
		jsr [a,x]
20		leau sizeof_mon,u
		cmpu #monsters_end
		blo 10B
		rts

		DATA
jtbl_monster_state
		STATE_JTBL
		fdb mon_centred
		fdb mon_check_direction
		fdb spr_move1
		fdb spr_delay
		fdb spr_move2
		STATE_ID "state_mon_landed"
		fdb mon_landed
		STATE_ID "state_mon_new_room"
		fdb spr_delay
		fdb spr_delay
		fdb spr_delay
		STATE_ID "state_mon_landed2"
		fdb mon_landed
		; keep these as the last states
		STATE_ID "state_mon_die"
		fdb mon_die
		fdb mon_dying
		CODE

; Players.

process_players
		lda level
		cmpa #24
		bne 10F
		; special-case finishing level 25
		lda playing
		eora #$0f
		sta finished
10		ldu #players
20		lda spr_state,u
		bmi 30F
		lda spr_state,u
		ldx #jtbl_player_state
		jsr [a,x]
30		leau sizeof_plr,u
		cmpu #players_end
		blo 20B
		rts

		DATA
jtbl_player_state
		STATE_JTBL
		STATE_ID "state_plr_centred"
		fdb plr_centred
		fdb plr_delay
		fdb plr_delay
		fdb plr_delay
		fdb plr_landed_nu
		STATE_ID "state_plr_move1"
		fdb spr_move1
		STATE_ID "state_plr_move2"
		fdb spr_move2
		STATE_ID "state_plr_landed"
		fdb plr_landed
		STATE_ID "state_plr_draw_room"
		fdb plr_draw_room
		STATE_ID "state_plr_draw_objects"
		fdb plr_draw_objects
		; keep these as the last states
		STATE_ID "state_plr_inactive"
		fdb plr_inactive
		STATE_ID "state_plr_exit0"
		fdb plr_exit0
		STATE_ID "state_plr_exitdelay"
		fdb spr_delay
		fdb plr_exit1
		STATE_ID "state_plr_die"
		fdb plr_die
		STATE_ID "state_plr_dying_delay"
		fdb spr_delay
		STATE_ID "state_plr_dying"
		fdb plr_dying
		STATE_ID "state_plr_undying_delay"
		fdb spr_delay
		STATE_ID "state_plr_undying"
		fdb plr_undying
		CODE

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Monster state handlers.

; Find next empty monster slot (if any) and create a new one.  Successive
; calls iterate through each trapdoor in turn.

new_monster
next_td		equ *+2
		ldy #trapdoors
		leau sizeof_obj,y
		cmpu #trapdoors_end
		blo 10F
		ldu #trapdoors
10		stu next_td
		ldu #monsters
20		lda obj_y,u
		deca
		bvs 30F
		leau sizeof_mon,u
		cmpu #monsters_end
		blo 20B
		rts
30		lda level		; XXX monster strength by level?
		lsra
mod_monster_hp	equ *+1
		adda #8
		sta spr_hp,u
		ldd obj_y,y
		std obj_y,u
		stb spr_cx,u
		jsr yx_to_woff
		stx obj_woff,u
		stx spr_owoff,u
		; fall through to mon_change_direction

; Choose next "random" direction for a monster to head in.  Actually, it
; simply cycles through the directions, but as it may be called for
; multiple monsters each in turn, can appear random.

mon_change_direction
next_dir	equ *+1
		lda #$06
10		adda #2
		anda #$06
		cmpa spr_facing,u	; same direction?
		beq 10B			; try again
		sta next_dir
		sta spr_facing,u
		clr spr_state,u
		; fall through to spr_set_tile

; Determine which tile to use for a sprite.  Calculates based on facing, x
; and y and adds to tilebase.  Because x & y vary, this animates sprite
; as it moves.

spr_set_tile
		lda spr_facing,u
		ldb obj_y,u
		eorb obj_x,u
		lsrb
		adca spr_tilebase,u
		sta obj_tile,u
		rts

mon_check_direction
		bsr spr_move
		bne 60F
		pshs u
		ldu #players
10		cmpd obj_y,u
		beq 40F			; monster hit player
		leau sizeof_plr,u
		cmpu #players_end
		blo 10B
		ldu #monsters
20		cmpd obj_y,u
		beq 50F
		leau sizeof_mon,u
		cmpu #monsters_end
		blo 20B
		; no collisions - monster is moving
		puls u
		std spr_next_y,u
		eora obj_y,u
		anda #$f8
		beq 30F
		bsr spr_set_tile
		lda #state_mon_new_room
		sta spr_state,u
		rts
30		bsr spr_set_tile
		jmp spr_inc_state
		; monster hit player
40		lda plr_bit,u
		bita dying
		bne 50F
		; no hit sound in GMC version
mod_mon_hit_player	equ *+1
		lda #$90		; decrement hp by 10-armour
		adda plr_armour,u
		jsr plr_damage
		; collision - change direction
50		puls u
60		bsr mon_change_direction
		jmp obj_draw

; Test moving a sprite based on 'facing'.  Belongs with other
; sprite-generic functions, but siting here saves some space.

; Entry:
; 	A = pre-loaded 'facing' (spr_move_a only)
; 	U -> sprite
; Returns:
; 	A,B = new y,x
; 	X -> stone bitmap byte
; 	CC.Z = 1 if not stone/door

; WARNING: called from player key scanning routine, which sets DP=$ff.  Ok
; for now as no DP vars are accessed, but be careful modifying.

spr_move
		lda spr_facing,u
spr_move_a
		ldx #jtbl_spr_move
		jmp [a,x]
		DATA
jtbl_spr_move	fdb spr_move_up,spr_move_down
		fdb spr_move_left,spr_move_right
		CODE
spr_move_up	ldd obj_y,u
		deca
		bra 10F
spr_move_down	ldd obj_y,u
		inca
10		ldx #tbl_x_to_bit
		pshs b
		ldb b,x
		ldx #bmp_wall+128
		bitb a,x
		puls b,pc
spr_move_left	ldd obj_y,u
		decb
		bpl 10B
		ldb #$07
		suba #48
		bra 10B
spr_move_right	ldd obj_y,u
		incb
		andb #7
		bne 10B
		adda #48
		bra 10B

mon_landed
		jsr spr_undraw
		ldd spr_next_y,u
		std obj_y,u
		stb spr_cx,u
		jsr yx_to_woff
		stx obj_woff,u
		stx spr_owoff,u
		clr spr_state,u
		; fall through to mon_centred

mon_centred
		jsr spr_set_tile
		bsr obj_draw
		bra spr_inc_state

mon_die
		lda #bones0
		bsr 10F
		bra spr_inc_state
mon_dying
		lda obj_tile,u
		beq 20F
		inca
		cmpa #bones2
		bhi 20F
10		sta obj_tile,u
		ldx obj_woff,u
		; use cx to position the death anim exactly where the
		; monster was
		ldb spr_cx,u
		bra obj_draw_vx
20		jsr spr_undraw
		ldd spr_next_y,u
		cmpd obj_y,u
		beq 30F
		std obj_y,u
		jsr yx_to_woff
		jsr spr_undraw_v
30		clr obj_tile,u
		ldb #$80
		stb obj_y,u
		rts

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Shared sprite routines.  Common to monsters & players.

		DATA
		; offset,tileset select
tbl_obj_woff1	fcb -3*fb_w,0	; up, x&1=0
		fcb -3*fb_w,1	; up, x&1=1
		fcb 3*fb_w,0	; down, x&1=0
		fcb 3*fb_w,1	; down, x&1=1
		fcb -1,1	; left, x&1=0
		fcb 0,0		; left, x&1=1
		fcb 0,1		; right, x&1=0
		fcb 1,0		; right, x&1=1
		; offset only (tileset remains)
		; could just reuse woff1 table and invert x&1
tbl_obj_woff2	fcb -3*fb_w	; up, x&1=0
		fcb -3*fb_w	; up, x&1=1
		fcb 3*fb_w	; down, x&1=0
		fcb 3*fb_w	; down, x&1=1
		fcb 0		; left, x&1=0
		fcb -1		; left, x&1=1
		fcb 1		; right, x&1=0
		fcb 0		; right, x&1=1
		CODE

spr_move1
		jsr spr_undraw
		ldx #tbl_obj_woff1
		ldb obj_x,u
		andb #1
		orb spr_facing,u
		lslb
		ldd b,x
		; A = offset from current screen position
		; B = tile set select (0=A, 1=B)
		ldx spr_owoff,u
		leax a,x
		stx obj_woff,u
		stb spr_cx,u
		bsr obj_draw_vx
		bra spr_inc_state

spr_inc_state	lda spr_state,u
spr_delay	adda #2
		sta spr_state,u
spr_nop
99		rts

spr_move2
		jsr spr_undraw
		ldx #tbl_obj_woff2
		ldb obj_x,u
		stb spr_cx,u
		andb #1
		orb spr_facing,u
		lda b,x
		; A = offset from current screen position
		; B = tile set select (0=A, 1=B)
		ldx obj_woff,u
		leax a,x
		stx obj_woff,u
		bsr obj_draw_v
		bra spr_inc_state

spr_draw_cx
		ldx obj_woff,u
		ldb spr_cx,u
		bra obj_draw_vx

; Draw object.
; Entry:
; 	U -> object
obj_draw	ldx obj_woff,u
		; XXX this is accounting for the fudged shot delay
		; is there a better way to do this?
		bmi 99B
; ... if X already == woff
obj_draw_v	ldb obj_x,u
; ... if B *also* already == x
obj_draw_vx	lda obj_y,u
		anda #$f8
		sta tmp0		; preserve room
		lda obj_tile,u
; ... if A *also* already == tile, but note must manually set tmp0=yroom
; don't need U to point to an object to call this
obj_draw_vxt	stx tmp1		; preserve woff
obj_draw_vxt1	pshs u
		lsrb
		bcs 10F
		ldx #tbl_tiles_a
		ldy #single_tile_x_a
		bra 20F
10		ldx #tbl_tiles_b
		ldy #single_tile_x_b
20		leax a,x
		ldu a,x
		pshs u
		lda tmp0		; re-fetch room
		cmpa player1+plr_yroom
		bne 30F
		ldx tmp1		; re-fetch woff
		leax p1_win+128,x
		jsr ,y
		lda tmp0		; re-fetch room
30		cmpa player2+plr_yroom
		bne 40F
		ldx tmp1		; re-fetch woff
		leax p2_win+128,x
		ldu ,s
		jsr ,y
		lda tmp0		; re-fetch room
40		cmpa player3+plr_yroom
		bne 50F
		ldx tmp1		; re-fetch woff
		leax p3_win+128,x
		ldu ,s
		jsr ,y
		lda tmp0		; re-fetch room
50		cmpa player4+plr_yroom
		bne 90F
		ldx tmp1		; re-fetch woff
		leax p4_win+128,x
		ldu ,s
		jsr ,y
90		puls x,u,pc

; Undraw sprite.  Sprite's current map position isn't updated until
; movement finishes ("lands"), so during movement, this will only redraw
; the tile the sprite is moving *from*.
; Entry:
; 	U -> sprite

spr_undraw
		ldx spr_owoff,u		; original (centred) woff
spr_undraw_v
		stx tmp1		; preserve woff
		jsr obj_test_pickup
		beq 30F
		ldy #pickups
		ldx obj_y,u
10		lda obj_tile,y
		cmpa #always_draw_tile	; this tile always drawn?
		bhs 20F			; skip - drawn soon enough anyway
		cmpx obj_y,y
		beq 40F
20		leay sizeof_obj,y
		cmpy #objects_end
		blo 10B
		rts
30		clra
40		ldb obj_y,u
		andb #$f8
		stb tmp0
		ldb obj_x,u
		jmp obj_draw_vxt1

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Player state handlers.

; Second state while changing room.  Draws all objects once.
; Subsequently, keys and animated objects are redrawn, anything else is
; only redrawn as part of undrawing a sprite.

plr_draw_objects
		bsr spr_undraw		; from other windows
		ldd plr_win,u
		std 30F
		ldb spr_next_y,u
		stb 20F
		pshs u
mod_draw_chalice_1
		jsr plr_update_chalice_b
		; draw objects
		ldy #objects
10		lda obj_tile,y
		beq 40F
		ldb obj_y,y
20		equ *+1
		eorb #$00
		andb #$f8
		bne 40F
		ldx obj_woff,y
30		equ *+2
		leax >$0000,x
		jsr draw_object_once
40		leay sizeof_obj,y
		cmpy #objects_end
		blo 10B
		puls u
		lda plr_bit,u
		bita playing
		beq plr_set_inactive
		ldb #state_plr_landed
		stb spr_state,u
		bsr plr_update_pos
		lda plr_bit,u
		bita dying
		lbne plr_kill
		rts

; Still need to update yroom for inactive players, as their windows get
; updated anyway.  However, player setup invalidates their (y,x) and the
; inactive state doesn't update this, so checks against their position
; will fail.

plr_set_inactive
		lda #state_plr_inactive
		sta spr_state,u
		lda spr_next_y,u
		anda #$f8
		sta plr_yroom,u
		rts

plr_update_pos	ldd spr_next_y,u
		std obj_y,u
		anda #$f8
		sta plr_yroom,u
		ldd obj_y,u
		stb spr_cx,u
		bsr yx_to_woff
		stx obj_woff,u
		stx spr_owoff,u
		rts

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Entry:
; 	A,B = y,x
; Returns:
; 	X = window offset
yx_to_woff
		pshs a,b,u
		ldu #tbl_y_to_woff
		anda #7
		lsla
		ldx a,u
		ldu #tbl_x_to_woff
		lda b,u
		leax a,x
		puls a,b,u,pc

		DATA
		; vertical offsets within a window
tbl_y_to_woff	fdb $0000,$0120
		fdb $0240,$0360
		fdb $0480,$05a0
		fdb $06c0,$07e0
		; horizontal offsets within a window, and within the whole
		; screen (when writing menus, messages, etc.)
tbl_x_to_woff	fcb 0,1,3,4
		fcb 6,7,9,10
		fcb 12,13,15,16
		fcb 18,19,21,22
		fcb 24,25,27,28
		fcb 30
		CODE

		; - - - - - - - - - - - - - - - - - - - - - - -

; Entry:
; 	U -> sprite
; Returns:
; 	A = y
; 	B = x bit
; 	X -> pickup bitmap byte
; 	CC.Z = 1 if no pickup
obj_test_pickup
		ldd obj_y,u
obj_test_pickup_d
		ldx #tbl_x_to_bit
		ldb b,x
		ldx #bmp_objects+128
		bitb a,x
		rts

		; - - - - - - - - - - - - - - - - - - - - - - -

plr_landed
		jsr spr_undraw
plr_landed_nu
		bsr plr_update_pos
		clr spr_state,u
		bsr obj_test_pickup_d
		beq 90F
		; pickups...
		ldx obj_y,u
		ldy #pickups
10		lda obj_tile,y
		beq 30F
		bmi 30F
		cmpx obj_y,y
		bne 30F
		pshs a,x,y
		; pickups MUST have 0 < obj_tile <= $3f
		ldx #jtbl_pickup
		lsla
		jsr [a,x]
		puls a
		ldx #tbl_pickup_scores
		lda a,x
		beq 20F
		ldy 2,s
		jsr obj_clr_pickup
		clr obj_tile,y
		jsr plr_add_score
20		puls x,y
30		leay sizeof_obj,y
		cmpy #pickups_end
		blo 10B
		; fall through to plr_centred
90

plr_centred
		jsr spr_set_tile
		lda plr_control,u
		beq 50F
plr_test_fire	bpl plr_fire

		anda #$7f
		sta spr_facing,u
		jsr spr_set_tile	; yes, again
		lda dying
		sta tmp0
		jsr spr_move
		beq 10F
mod_check_chalice
		jsr check_chalice
		cmpd plr_key_opens,u
		bne 50F
		jmp plr_open_door
10		lsr tmp0
		bcs 20F
		cmpd player1+obj_y
		beq 50F
20		lsr tmp0
		bcs 30F
		cmpd player2+obj_y
		beq 50F
30		lsr tmp0
		bcs 40F
		cmpd player3+obj_y
		beq 50F
40		lsr tmp0
		bcs 60F
		cmpd player4+obj_y
		bne 60F
50		jsr obj_draw
		jmp spr_inc_state

60		ldy #monsters
70		cmpd obj_y,y
		beq plr_bump
		leay sizeof_mon,y
		cmpy #monsters_end
		blo 70B

		std spr_next_y,u
		anda #$f8
		cmpa plr_yroom,u
		beq 80F
		dec plr_yroom,u	; invalidate yroom
		lda #state_plr_draw_room
		sta spr_state,u
		rts
80		jsr obj_draw
		ldb #state_plr_move1
		lda plr_speed,u
		anda #$7f
		beq 90F
		ldb #state_plr_landed
90		stb spr_state,u
		bra plr_move1

plr_delay
		lda plr_control,u
		bne plr_test_fire
		jmp spr_inc_state

; Entry:
; 	A = fire control (1=>fire, 2=>magic)
; 	U -> player
plr_fire
		lsra
		bne plr_magic
10		lda plr_nshots,u
		cmpa plr_ammo,u
		bhs 50B
		inca
		sta plr_nshots,u
		lda obj_tile,u
		eora #$01
		sta obj_tile,u
		ldy #shots
20		lda obj_tile,y
		beq 30F
		leay sizeof_obj,y
		bra 20B		; XXX *should* be safe...
30		ldb spr_facing,u
		stb obj_shot_facing,y
		lsrb
		addb #8
		addb spr_tilebase,u
		stb obj_tile,y
		ldd obj_y,u
		std obj_y,y
		ldd #$fffe	; 1 frame delay
		std obj_woff,y
		stu obj_shot_plr,y

		ldx #snd_fire
		stx psg_latch_c3

		jmp obj_draw	; draw player

		; player bumped into monster
		; limit amount of damage caused by only dropping health if
		; monster is "landed"
plr_bump	lda spr_state,y
		cmpa #state_mon_landed
		beq 10F
		cmpa #state_mon_landed2
		beq 10F
		rts
10		; no bump sound in GMC version
		lda #$99		; decrement hp by 1
		jmp plr_damage

		; Continues on from plr_centred, decided player is moving,
		; no speedyboots.  Hack to store decision for
		; joystick-controlled players.
plr_move1	ldb spr_facing,u
		andb #$04	; only care about axis
		lda plr_bit,u
		cmpa #$04
		bne 10F
		stb p3_moved
		bra 99F
10		cmpa #$08
		bne 99F
		stb p4_moved
99		rts

plr_magic
		jsr obj_draw
		lda #$94	; -6HP for player 1
		ldb plr_bit,u
		lsrb
		beq 10F
		lda #$97	; -3HP for player 2
10		jsr plr_damage
		; only kill monsters in player's window
		lda obj_y,u
		sta tmp0+1
		; calculate score and mark monsters for death
		ldy #monsters
20		lda obj_y,y
		eora tmp0+1
		anda #$f8
		bne 30F		; only monsters in the right room
		ldd #$0100|state_mon_die
		stb spr_state,y
		jsr plr_add_score
30		leay sizeof_mon,y
		cmpy #monsters_end
		blo 20B
		ldx #snd_magic
		stx psg_latch_c2
		jsr psg_latch_all
		; step monsters through death and play magic sound fx
		bsr magic_mon_step
		bsr magic_mon_step
		bsr magic_mon_step
		; fall through to magic_mon_step

magic_mon_step	pshs u
		bsr mon_step
		; funny sound
		bsr magic_playfrag
		bsr magic_playfrag
		bsr magic_playfrag
		bsr magic_playfrag
		puls u,pc

magic_playfrag	ldx #1280
10		leax -1,x
		bne 10B
		jmp psg_playfrag

mon_step	pshs u
		ldu #monsters
10		lda obj_y,u
		eora tmp0+1
		anda #$f8
		bne 20F		; only monsters in the right room
		lda spr_state,u
		ldx #jtbl_monster_state
		jsr [a,x]
20		leau sizeof_mon,u
		cmpu #monsters_end
		blo 10B
		puls u,pc

; Entry:
; 	A,B = y,x of held key = intended player destination
; 	U -> player
plr_open_door
		ldx #doors
10		cmpd obj_y,x
		beq 20F
		leax sizeof_obj,x
		bra 10B		; XXX should be safe loop
20		clr obj_tile,x
		pshs a,b,u
		leau ,x
		jsr obj_draw
		puls a,b,u
		; clear appropriate bit in wall bitmap
		ldx #tbl_x_to_bit
		ldb b,x
		ldx #bmp_wall+128
		comb
		andb a,x
		stb a,x
		; invalidate player's held key
		ldd #$8080
		std plr_key_opens,u
		ldx #snd_door
		stx psg_latch_c1
		lda #7
		jsr plr_add_score
		; consider setting state such that there's a slight delay
		; before moving on?
		; fall through to plr_undraw_key

plr_undraw_key	ldx plr_swin,u
		leax 2+32,x
PATCH_undraw_key	equ *+1
		ldd #$aaaa
		sta -32,x
		stb ,x
		sta 32,x
plr_inactive	rts

		; if player had started to move to another square
		; (spr_next_y != obj_y), undraw that square
plr_die		ldd spr_next_y,u
		cmpd obj_y,u
		beq 10F
		ldx obj_y,u
		pshs x			; preserve old y,x
		std obj_y,u
		jsr yx_to_woff
		jsr spr_undraw_v
		puls x
		stx obj_y,u		; restore old y,x
		stx spr_next_y,u	; ensure resurrects in same place
10		lda #bones0-1
		sta obj_tile,u
		jmp spr_inc_state

plr_dying	lda #state_plr_dying_delay
		sta spr_state,u
		lda obj_tile,u
		cmpa #bones7
		blo 20F
		ldb plr_bit,u
		bitb dying
		bne 10F
		ldb #state_plr_undying
		stb spr_state,u
		bra 30F
10		orb dead
		stb dead
		fcb $21		; brn - skip inca
20		inca
30		sta obj_tile,u
		ldx spr_owoff,u
		; don't use cx - player should die centred
		ldb obj_x,u
		jmp obj_draw_vx

; Resurrection runs the death animation in reverse.
plr_undying
		lda #state_plr_undying_delay
		sta spr_state,u
		lda obj_tile,u
		cmpa #bones0
		bls 10F
		deca
		bra 30B
10 		lda #state_plr_landed
		sta spr_state,u
		lda #$32
		sta spr_hp,u
		jmp plr_draw_hp

; Test if player is moving into the chalice, and if so set that player's
; bit in 'finished'.  When 'finished' contains all living players, the
; game is won.

; It's ok to destroy A here, as B will still test false in plr_centred,
; and there are no actual doors on level 25.
check_chalice	cmpa #$14
		blo 10F
		cmpa #$15
		bhi 10F
		cmpb #2
		blo 10F
		cmpb #3
		bhi 10F
		lda plr_bit,u
		ora finished
		bra 20F
10		lda plr_bit,u
		coma
		anda finished
20		sta finished
		rts

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Handle player damage and death.  Death may involve dropping a held key,
; so that's here too.

; Sets bit in pickup table.
; Entry:
; 	A = y
; 	B = x
set_pickup
		pshs b,x
		ldx #tbl_x_to_bit
		ldb b,x
		ldx #bmp_objects+128
		orb a,x
		stb a,x
		puls b,x,pc

; Entry:
; 	A = number to add to player hp (100-damage in BCD)
; 	U -> player
plr_damage	adda spr_hp,u
		daa
		bcs 10F
		bsr plr_kill
		clra
10		sta spr_hp,u
		jmp plr_draw_hp

; Entry:
; 	U -> player
plr_kill	lda spr_state,u
		cmpa #state_plr_inactive
		bhs 99F
		lda dying
		ora plr_bit,u
		sta dying
		lda spr_state,u
		cmpa #state_plr_draw_room
		blo 10F
		cmpa #state_plr_draw_objects
		bhi 10F
99		rts
10		ldx death_fifo_end
		stu ,x++
		stx death_fifo_end
		jsr kill_speed
		lda #state_plr_die
		sta spr_state,u
		ldd obj_y,u
		; fall through to drop_key

; Drop key held by player in specified location.  Checks player has a key
; in the first place.

; Entry:
; 	A,B = y,x
; 	U -> player
drop_key
		pshs a,b,y
		lda plr_key_opens,u
		deca
		bvs 90F		; y = $80 - no key!
		jsr plr_undraw_key
		ldy #keys
10		lda obj_tile,y
		beq 20F
		leay sizeof_obj,y
		bra 10B
20		ldd plr_key_opens,u
		std obj_key_opens,y
		clra
		ldb plr_speed,u
		lslb
		adca #key
		sta obj_tile,y
		ldd ,s
		std obj_y,y
		jsr yx_to_woff
		stx obj_woff,y
		bsr set_pickup
		ldd #$8080
		std plr_key_opens,u
90		puls a,b,y,pc

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Shot processing

shot_hit_player
		ldx obj_shot_plr,u	; X -> player shot belongs to
		pshs u
		leau ,y			; U -> player being shot
		lda #$9a		; decrement hp by shot power
		suba plr_power,x
		bsr plr_damage
		puls u
		; fall through to shot_remove

shot_remove	clr obj_tile,u
		ldx obj_shot_plr,u
		dec plr_nshots,x
		bra process_next_shot

process_shots	ldu #shots
10		lda obj_tile,u
		beq process_next_shot
		ldx obj_woff,u
		bpl 12F
		leax 1,x
		beq 14F
		stx obj_woff,u
		bra process_next_shot
12		jsr spr_undraw_v
14		jsr spr_move
		bne shot_remove

		; test new shot position against drainers.
20		tfr d,x
		ldy #items
22		lda obj_tile,y
		cmpa #drainer
		bne 24F
		cmpx obj_y,y
		beq shot_hit_drainer
24		leay sizeof_obj,y
		cmpy #items_end
		blo 22B

		; test new shot position against players.  dying or dead
		; players still block shots.
test_shot_player
		ldy #players
30		cmpx obj_y,y
		beq shot_hit_player
		leay sizeof_plr,y
		cmpy #players_end
		blo 30B

		; test new shot position against monsters.  dying monsters
		; do NOT block shots, so test for this.
		ldy #monsters
40		cmpx obj_y,y
		bne 42F
		lda spr_state,y
		cmpa #state_mon_die
		blo shot_hit_monster
42		leay sizeof_mon,y
		cmpy #monsters_end
		blo 40B

		; all tests done, store new position, and calculate new
		; window offset.
		stx obj_y,u
		tfr x,d
		jsr yx_to_woff
		stx obj_woff,u

process_next_shot
		leau sizeof_obj,u
		cmpu #shots_end
		blo 10B
		rts

shot_hit_drainer
		; invert direction of shot
		lda obj_shot_facing,u
		eora #$02
		sta obj_shot_facing,u
		lda obj_tile,u
		eora #$01
		sta obj_tile,u
		; increment player's score
		pshs x,u
		ldu obj_shot_plr,u
		lda #1
		jsr plr_add_score
		puls x,u
		; decrement drainer "hp"
		dec obj_drainer_hp,y
		; if not "dead", return to check against players
		; this makes bones block reflection of shots
		bne test_shot_player
		; disable drainer in object table
		clr obj_tile,y
		; remove drainer from pickup table
		jsr check_clr_pickup
		; and return to check against players
		bra test_shot_player

shot_hit_monster
		ldx #snd_monster_hit
		stx psg_latch_c3
		lda #$9a
		pshs u
		ldu obj_shot_plr,u
		lda spr_hp,y
		suba plr_power,u
		bcc 10F
		ldd #$0100|state_mon_die	; 1pt for monster
		stb spr_state,y
		jsr plr_add_score
		puls u
		jmp shot_remove
10		sta spr_hp,y
		puls u
		jmp shot_remove

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Pickup dispatch routines.

; Entry:
; 	A = object type
; 	Y -> object
; 	U -> player

pickup_potion
		bsr sound_low
		lda plr_score0,u
		anda #$0f
		ldb #$11
		mul
10		stb spr_hp,u
		jmp plr_draw_hp

pickup_food
		ldx #snd_food
		stx psg_latch_c1
		lda spr_hp,u
		adda #$10
		daa
		bcc 10F
		lda #$99
10		sta spr_hp,u
		jsr plr_draw_hp
		bra kill_speed

pickup_armour	bsr sound_low
		lda plr_armour,u
		inca
		cmpa #7
		bhi 99F
		sta plr_armour,u
		; fall through to plr_draw_armour

plr_draw_armour	ldx plr_swin,u
		leax 2*fb_w+3+128,x
		lda plr_armour,u
		bra draw_snum

pickup_power	bsr sound_low
		lda plr_power,u
		inca
		cmpa #9
		bhi 99F
		sta plr_power,u
		; fall through to plr_draw_power

plr_draw_power	ldx plr_swin,u
		leax 10*fb_w+2+128,x
		lda plr_power,u
		; fall through to draw_snum

draw_snum	pshs u
		ldu snum_tiles_start
		ldb #5
		mul
		leau b,u
		pulu d
		sta -128,x
		stb -96,x
		pulu d
		sta -64,x
		stb -32,x
		lda ,u
		sta ,x+
		puls u,pc

pickup_weapon
		bsr sound_low
		lda plr_ammo,u
		inca
		cmpa #3
		bhi 99F
		sta plr_ammo,u
		jmp plr_draw_ammo

sound_low
		ldx #snd_low
		stx psg_latch_c3
		; fall through to kill_speed

kill_speed	clra
		lsl plr_speed,u
		rora
		sta plr_speed,u
		; fall through to plr_undraw_speed

plr_undraw_speed
		ldx plr_swin,u
		leax 4*fb_w+2+64,x
PATCH_undraw_speed	equ *+1
		ldd #$aaaa
		sta -64,x
		stb -32,x
		sta ,x
		stb 32,x
		sta 64,x
pickup_nop
99		rts

pickup_money
		ldx #snd_money
		stx psg_latch_c2
		bra kill_speed

pickup_cross
		bsr sound_low
		lda dying
		beq 99F
		ldx death_fifo_end
		leax -2,x
		stx death_fifo_end
		ldx death_fifo
		ldd death_fifo+2
		std death_fifo
		ldd death_fifo+4
		std death_fifo+2
		ldd death_fifo+6
		std death_fifo+4
		; unset dying and dead
		lda plr_bit,x
		coma
		tfr a,b
		anda dying
		sta dying
		andb dead
		stb dead
99		rts

pickup_speed
		bsr sound_low
		; this is safe!
		inc plr_speed,u
		; fall through to plr_draw_speed

plr_draw_speed	ldx plr_swin,u
		leax 4*fb_w+2+64,x
PATCH_draw_speed0	equ *+1
		ldd #$a69a
		sta -64,x
		stb -32,x
		sta ,x
		stb 32,x
PATCH_draw_speed1	equ *+1
		lda #$6a
		sta 64,x
		rts

pickup_key
		pshs a			; preserve key type

		; no sound priorities in GMC version, so hearing key
		; juggling very much depends on nothing else using the
		; channel.
		ldx #snd_key
		stx psg_latch_c2

		ldd plr_key_opens,u
		ldx obj_key_opens,y
		stx plr_key_opens,u
		std obj_key_opens,y

		clra
		ldb plr_speed,u		; top bit of plr_speed is key type
		lslb
		adca #key		; tile is #key or #key+1
		sta obj_tile,y		; update object tile

		puls a			; which key type was picked up?
		lsra			; tile id was shifted left
		lsra			; and now we want lowest bit
		rorb			; into top bit of plr_speed
		stb plr_speed,u

		; clear object tile if player wasn't actually holding a
		; key.  means some of above was redundant...
		ldd obj_key_opens,y
		cmpd #$8080
		bne 10F
		clr obj_tile,y
		ldx obj_y,u
		bsr check_clr_pickup
10		; fall through to plr_draw_key

plr_draw_key	ldx plr_swin,u
		leax 2+32,x
		lda plr_speed,u
		bmi plr_draw_key1
PATCH_draw_key0	equ *+1
		ldd #$a656
		sta -32,x
		stb ,x
PATCH_draw_key1	equ *+1
		lda #$66
		sta 32,x
		rts

plr_draw_key1
PATCH_draw_key01	equ *+1
		ldd #$a202
		sta -32,x
		stb ,x
PATCH_draw_key11	equ *+1
		lda #$22
		sta 32,x
99		rts

; Clear bitmap entry for object y.
; Entry:
; 	Y -> object
obj_clr_pickup	pshs a,x,y
		ldd obj_y,y
		bra clr_pickup_d_

; Clear bitmap entry for y,x (held in x) if no other pickups there.  Only
; necessary to check like this for keys and drainers.
; Entry:
; 	X = y,x
check_clr_pickup
		pshs a,x,y
		ldy #pickups
10		lda obj_tile,y
		beq 20F
		cmpx obj_y,y
		beq 30F
20		leay sizeof_obj,y
		cmpy #pickups_end
		blo 10B
clr_pickup_x_	tfr x,d
clr_pickup_d_	ldx #tbl_x_to_bit
		ldb b,x
		ldx #bmp_objects+128
		comb
		andb a,x
		stb a,x
30		puls a,x,y,pc

; Clear bitmap entry for y,x (held in x) without checking.
; Entry:
; 	X = y,x
clr_pickup_x	pshs a,x,y
		bra clr_pickup_x_

		DATA
tbl_x_to_bit	fcb $80,$40,$20,$10
		fcb $08,$04,$02,$01
		CODE

pickup_exit
		lda exiting
		bne 99B
		lda #facing_up
		jsr spr_move_a
		beq 10F
		lda #facing_right
		jsr spr_move_a
		beq 10F
		lda #facing_down
		jsr spr_move_a
		beq 10F
		lda #facing_left
		jsr spr_move_a
		; XXX assuming one will work
10		jsr drop_key
		lda #state_plr_exit0
		sta spr_state,u
		sta exiting
		; invalidate x, but store its value in cx for use drawing
		; exit sprites
		lda obj_x,u
		sta spr_cx,u
		com obj_x,u
		ldd #tbl_exit_data
		std spr_tmp0,u
		jsr kill_speed
		ldx #snd_exit
		stx psg_latch_c2
		; skip return address, stacked object type and return:
		bra 90F

		; not a pickup: teleport player
pickup_tport
		ldx #snd_tport
		stx psg_latch_c2
10		leay sizeof_obj,y
		cmpy #items_end
		blo 20F
		ldy #items
20		lda obj_tile,y
		cmpa #tport
		bne 10B
		ldd obj_y,y
		deca
		std spr_next_y,u
		anda #$f8
		cmpa plr_yroom,u
		beq 30F
		dec plr_yroom,u	; invalidate yroom
		lda #state_plr_draw_room
		bra 40F
30		lda #state_plr_landed
40		sta spr_state,u
89		jsr kill_speed	; XXX test this
90		leas 3,s	; skip return address, stacked object type
		puls x,y,pc

		; not a pickup: drain health
pickup_drainer
		ldx #snd_drainer
		stx psg_latch_c3
mod_drainer_dmg	equ *+1
		lda #$80	; -20HP
		jsr plr_damage
		; kill speed, skip return address, stacked object type, return
		bra 89B

plr_exit1
		lda #state_plr_exitdelay-2
		sta spr_state,u
		; fall through to plr_exit0

plr_exit0
		ldx spr_tmp0,u
		cmpx #tbl_exit_data_end
		bhs 10F
		lda ,x+
		adda spr_tilebase,u
		sta obj_tile,u
		stx spr_tmp0,u
		jsr spr_draw_cx
		jmp spr_inc_state

10
		clr exiting
		lda finished
		ora plr_bit,u
		sta finished
		lda #state_plr_inactive
		sta spr_state,u
		lda #exit
		sta obj_tile,u
		jmp spr_draw_cx

		DATA

tbl_exit_data	fcb p1_up0-p1_tilebase
		fcb p1_right0-p1_tilebase
		fcb p1_down0-p1_tilebase
		fcb p1_left0-p1_tilebase
		fcb p1_exit4-p1_tilebase
		fcb p1_exit5-p1_tilebase
		fcb p1_exit6-p1_tilebase
		fcb p1_exit7-p1_tilebase
		fcb p1_exit8-p1_tilebase
		fcb p1_exit9-p1_tilebase
		fcb p1_exit10-p1_tilebase
		fcb p1_exit11-p1_tilebase
tbl_exit_data_end

tbl_pickup_scores	equ *-1		; floor not a pickup
		fcb 0		; exit
		fcb 0		; trapdoor
		fcb $05		; food
		fcb 0		; tport
		fcb $10		; armour
		fcb $10		; potion
		fcb $10		; weapon
		fcb $10		; cross
		fcb $10		; speed
		fcb 0		; drainer
		fcb $20		; money
		fcb $10		; power
		fcb 0		; dummy
		fcb 0		; key
		fcb 0		; key1

jtbl_pickup	equ *-2			; no dispatch for floor
		fdb pickup_exit
		fdb pickup_nop		; trapdoor
		fdb pickup_food
		fdb pickup_tport
		fdb pickup_armour
		fdb pickup_potion
		fdb pickup_weapon
		fdb pickup_cross
		fdb pickup_speed
		fdb pickup_drainer
		fdb pickup_money
		fdb pickup_power
		fdb pickup_key		; dummy
		fdb pickup_key
		fdb pickup_key		; key1

		CODE

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Wait for any key

wait_keypress
		; wait if key held
10		bsr firebutton_pressed
		bne 10B
		bsr wait_keyheld
		; wait until key pressed
10		lda reg_pia0_pdra
		ora #$80
		inca
		beq 10B
		bsr firebutton_pressed
		bne 10B
		rts

wait_keyheld
		; wait if key held
10		bsr test_keypressed
		bne 10B
		rts

test_keypressed
		clr reg_pia0_pdrb
		lda reg_pia0_pdra
		ora #$80
		inca
		rts

; For menus: scan table of KEYDSP_ENTRY and jump to handler

wait_keypress_jump
		bsr wait_keyheld
10		bsr test_keypress_jump
		bra 10B


test_keypress_jump
		; test firebutton before & after scan
20		bsr firebutton_pressed
		bne 99F
		leau -2,x
30		leau 4,u
		ldd -2,u
		beq 99F
		sta reg_pia0_pdrb
		bitb reg_pia0_pdra
		bne 30B
		bsr firebutton_pressed
		bne 99F
		leas 2,s
		pulu pc
99		rts

firebutton_pressed
		lda #$ff
		sta reg_pia0_pdrb
		lda reg_pia0_pdra
		ora #$80
		inca
99		rts

; Entry:
; 	U -> textlist
; textlist:
; 	[ y, x, &string ]..., 0

write_textlist
10		lda ,u+
		beq 99B
		deca
		pulu b,y
		bsr write_at
		bra 10B

; Entry:
; 	U -> charlist
; charlist:
; 	[ y, x, char ] ..., 0

write_charlist
10		lda ,u+
		beq 99B
		deca
		pulu b
		std text_cursor_y
		pulu a
		bsr write_char
		bra 10B

; Entry:
; 	Y -> y, x, string (write_screen)
; 	Y -> string (write_at)
; 	A, B = y, x (write_at)
; string:
; 	[ char ]..., 0

write_screen	ldd ,y++		; y,x
write_at	std text_cursor_y
		stb text_line_x
10		lda ,y+
		beq 99B
		cmpa #$7d	; Dunjunz logo
		beq write_logo
		cmpa #$7e	; Nudge left (kerning)
		beq 50F
		cmpa #$7f	; LF
		beq 40F
		bsr write_char
		bra 10B

write_logo	bsr get_write_x
		pshs u
		ldu logo
		lda #17
		pshs a
30		lda #15
32		ldb ,u+
		stb ,x+
		deca
		bne 32B
		leax fb_w-15,x
		dec ,s
		bne 30B
		puls a,u
		; fall through to LF

		; LF
40		ldd text_line_x
		sta text_cursor_x
mod_line_height	equ *+1
		addb #11
		stb text_cursor_y
		bra 10B

		; Nudge left (kerning)
50		dec text_cursor_x
		bra 10B

get_write_x	lda text_cursor_y
		ldb #32
		mul
		addd #fb0
		tfr d,x
		ldb text_cursor_x
		lsrb
		lsrb
		abx
		rts

write_char	pshs u
		sta text_descender	; top bit set indicates descender
		anda #$7f
		ldb #7
		mul
		addd #textfont
		tfr d,u
		pulu a
		sta text_char_width
		bsr get_write_x

		lda #6
		sta tmp2+1
10		lda ,u+
		bsr text_emit_byte
		dec tmp2+1
		bne 10B

		lda text_descender
		bpl text_char_next
		anda #$06
		ldu #jtbl_descender
		jmp [a,u]

text_emit_byte	bsr text_draw_byte
		leax fb_w,x
		rts
text_draw_byte	ldb text_cursor_x
		pshs b
		ldb #$ff
		lsr ,s
		bcc 10F
		rora
		rorb
		rora
		rorb
10		lsr ,s+
		bcc 20F
		rora
		rorb
		rora
		rorb
		rora
		rorb
		rora
		rorb
20		std text_tmp0
		coma
		comb
		anda text_colour
		andb text_colour
		std text_tmp1
		ldd ,x
		anda text_tmp0
		andb text_tmp0+1
		addd text_tmp1
		std ,x
		rts

descender_gy
		lda #~$0c
		bsr text_emit_byte
		lda #~$30
		bsr text_emit_byte
		lda #~$c0
		bsr text_draw_byte
		leax -1,x
		lda #~$ff
		bsr text_draw_byte
		leax -1,x
10		lda #~$03
20		bsr text_draw_byte
		; fall through to text_char_next

text_char_next
		lda text_cursor_x
		adda text_char_width
		sta text_cursor_x
		puls u,pc

descender_comma
		lda #~$c0
		bsr text_draw_byte
		leax fb_w-1,x
		bra 10B

descender_p
		lda #~$c0
		bsr text_emit_byte
		lda #~$c0
		bsr text_emit_byte
		lda #~$c0
		bra 20B

descender_q
		lda #~$0c
		bsr text_emit_byte
		lda #~$0c
		bsr text_emit_byte
		lda #~$0c
		bra 20B

		DATA
jtbl_descender	fdb descender_gy
		fdb descender_comma
		fdb descender_p
		fdb descender_q
		CODE

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		; CoCo 3 mode changes
set_cc3_palette
		pshs a,b
		lda #$44
		sta $ff90	; coco3 mode
		ldb #$09	; 32bytes per row, 4 colours
		lda $ff98	
		anda #$08	; preserve 50Hz/60Hz select
		ora #$80	; graphics, one line per row
		std $ff98
PATCH_cc3_c0c1	equ *+1
		ldd #$0000
		std $ffb0
PATCH_cc3_c2c3	equ *+1
		ldd #$0000
		std $ffb2
		sta $ff9a	; border == background
		ldd #$e040	; voff = $70200
		std $ff9d
		puls a,b,pc

; Unpack dzipped screen and draw player stats to it.

unpack_play_screen
		pshs x,y
		ldx play_screen_dz
		bra 10F
unpack_text_screen
		pshs x,y
		ldx text_screen_dz
10		lda vdg_mode
		sta reg_pia1_pdrb
		bsr set_cc3_palette

		lda #3
		sta reg_gmc_bank
		ldu #fb0
		ldy #fb0+178*fb_w
		sty dzip_end
		export PATCH_screen_border
PATCH_screen_border	equ *+1
		ldd #$0000
10		std ,u++
		std ,y++
		cmpu #fb0+14*fb_w
		blo 10B
		jsr ldunzip

		ldu #player1
10		jsr plr_draw_stats
		leau sizeof_plr,u
		cmpu #players_end
		blo 10B
		puls x,y,pc

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; dunzip - unzip simply coded zip data

; 1iiiiiii 0nnnnnnn - repeat 128-n (1-128) bytes from current + 1i (-ve 2s cmpl)
; 1iiiiiii 1jjjjjjj nnnnnnnn - repeat 128-n (1-256) bytes from current + 1ij
; 0nnnnnnn - directly copy next 128-n (1-128) bytes

; Entry:
; 	X -> zipped data
; 	U -> destination
; 	[dzip_end] = end of zipped data

ldunzip
ldunz_loop	ldd ,x++
		bpl ldunz_run	; run of 1-128 bytes
		tstb
		bpl ldunz_7_7
ldunz_14_8	lslb		; drop top bit of byte 2
		asra
		rorb		; asrd
		leay d,u
		ldb ,x+
		bra 10F		; copy 1-256 bytes (0 == 256)
ldunz_7_7	leay a,u	; copy 1-128 bytes
10		lda ,y+
		sta ,u+
		incb
		bvc 10B		; count UP until B == 128
		bra 30F
20		ldb ,x+
ldunz_run	stb ,u+
		inca
		bvc 20B		; count UP until B == 128
30		cmpu <dzip_end
		blo ldunz_loop
		rts

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		DATA

tile_map	
		; door1_v,door1_h,door_v,door_h
		; floor,exit,trapdoor,food
		; tport,armour,potion,weapon
		; cross,speed,drainer0,money0
		; power0,[power1],key,key1
		fcb 0,20
		; p1_up0,p1_up1,p1_down0,p1_down1
		; p1_left0,p1_left1,p1_right0,p1_right1
		; arrow_up,arrow_down,arrow_left,arrow_right
		; p1_exit4,p1_exit5,p1_exit6,p1_exit7
		fcb 20,16
		; exit8,exit9,exit10,exit11
		fcb 84,4
		; p2_up0,p2_up1,p2_down0,p2_down1
		; p2_left0,p2_left1,p2_right0,p2_right1
		; fball_up,fball_down,fball_left,fball_right
		; p2_exit4,p2_exit5,p2_exit6,p2_exit7
		fcb 36,16
		; exit8,exit9,exit10,exit11
		fcb 84,4
		; p3_up0,p3_up1,p3_down0,p3_down1
		; p3_left0,p3_left1,p3_right0,p3_right1
		; axe_up,axe_down,axe_left,axe_right
		; p3_exit4,p3_exit5,p3_exit6,p3_exit7
		fcb 52,16
		; exit8,exit9,exit10,exit11
		fcb 84,4
		; p4_up0,p4_up1,p4_down0,p4_down1
		; p4_left0,p4_left1,p4_right0,p4_right1
		; axe_up,axe_down,axe_left,axe_right
		; p4_exit4,p4_exit5,p4_exit6,p4_exit7
		; exit8,exit9,exit10,exit11
		; monster_up0,monster_up1,monster_down0,monster_down1
		; monster_left0,monster_left1,monster_right0,monster_right1
		; bones0,bones1,bones2,bones3
		; bones4,bones5,bones6,bones7
		fcb 68,36
		; drainer0,money0,power0
		fcb 14,3
		; money1,power1
		fcb 104,2

		fcb 0,0

		; yes,no,chalice_nw0_a
		fcb 0,3
		; chalice_nw1_a
		fcb 6,1
		; tile_stone01_a
		fcb 10,1

		fcb 0,0

		CODE

; Entry:
; 	X -> tile map (pairs of offset,count)
; 	Y -> tile pointers destination
; 	U -> tile data base
; Exit:
; 	Preserved: U
; 	X -> next tile map
; 	Y -> next tile pointers

populate_tile_ptrs
		pshs u

10		lda ,x+
		ldb #18
		mul
		ldu ,s
		leau d,u

		ldb ,x+
		bne 20F
		puls u,pc

20		stu ,y++
		leau 18,u
		decb
		bne 20B

		bra 10B

; Switch to the next video mode and update pointers to appropriate
; tiles/fonts/etc.

		export PATCH_max_video_mode

advance_video_mode
		ldb video_mode
		incb
PATCH_max_video_mode	equ *+1
		cmpb #3
		blo 10F
		clrb
10		stb video_mode
		lda #4
		mul
		ldx #video_mode_transitions
		abx
		ldu ,x
		pshs x

		ldx #tile_map
		ldy #tbl_tiles_a_start
		bsr populate_tile_ptrs
		leau 106*18,u

		ldx #tile_map
		;ldy #tbl_tiles_b_start
		bsr populate_tile_ptrs
		leau 106*18,u

		bsr populate_tile_ptrs
		leau 60*18,u

		ldb #10
		; populate tbl_lnum
10		stu ,y++
		leau 7,u
		decb
		bne 10B

		puls x

		ldu 2,x
		; fall through to apply_patchset

; Entry:
; 	U -> patchset
; See top of file for description of patchset.

		export apply_patchset

apply_patchset	pulu b,x	; B = count
		clra
10		leax a,x	; skip A bytes
20		pulu a
		sta ,x+
		decb
		bne 20B
		ldb ,u+		; B = 16-bit skip flag | count
		beq 99F		; 0 indicates finished
		bmi 30F		; bit 7 set indicates 16-bit skip
		pulu a		; A = 8-bit skip
		bra 10B
30		andb #$7f	; clear 16-bit skip flag
		pshs b
		pulu d
		leax d,x
		puls b		; B = count
		bra 20B
99		rts

		DATA

		export video_mode_transitions
video_mode_transitions
		fdb tiles_pal
		fdb ntsc1_pal_patchset		; ntsc1 -> pal patchset
		fdb tiles_ntsc0
		fdb pal_ntsc0_patchset		; pal -> ntsc0 patchset
		fdb tiles_ntsc1
		fdb ntsc0_ntsc1_patchset	; ntsc0 -> ntsc1 patchset

; PAL -> NTSC phase0 patchset

pal_ntsc0_patchset

		PATCH colour_bg
		fcb $00,$55,$ff,$aa
		ENDPATCH

		PATCH vdg_mode
		fcb $fa		; lda #$fa - VDG mode RG6, CSS1
		ENDPATCH

		PATCH logo
		fdb logo_ntsc
		ENDPATCH

		PATCH play_screen_dz
		fdb play_screen_ntsc_dz
		ENDPATCH

		PATCH text_screen_dz
		fdb text_screen_ntsc_dz
		ENDPATCH

		PATCH PATCH_draw_ammo0
		fcb $ff,$20	; ldd #$ff20
		ENDPATCH
		PATCH PATCH_draw_ammo1
		fcb $e8		; ldb #$e8
		ENDPATCH
		PATCH PATCH_draw_ammo2
		fcb $20		; ldb #$20
		ENDPATCH
		PATCH PATCH_draw_ammo3
		fcb $00,$00	; ldd #$0000
		ENDPATCH
		PATCH PATCH_draw_key0
		fcb $0c,$fc	; ldd #$0cfc
		ENDPATCH
		PATCH PATCH_draw_key1
		fcb $cc		; lda #$cc
		ENDPATCH
		PATCH PATCH_draw_key01
		fcb $04,$54	; ldd #$0454
		ENDPATCH
		PATCH PATCH_draw_key11
		fcb $44		; lda #$44
		ENDPATCH
		PATCH PATCH_undraw_key
		fcb $00,$00	; ldd #$0000
		ENDPATCH
		PATCH PATCH_draw_speed0
		fcb $0c,$30	; ldd #$0c30
		ENDPATCH
		PATCH PATCH_draw_speed1
		fcb $c0		; lda #$c0
		ENDPATCH
		PATCH PATCH_undraw_speed
		fcb $00,$00	; ldd #$0000
		ENDPATCH

		PATCH snum_tiles_start
		fdb snum_tiles_ntsc0-5
		ENDPATCH

		ENDPATCHSET

; NTSC phase0 -> NTSC phase1 patchset

ntsc0_ntsc1_patchset

		PATCH colour_bg
		fcb $00,$aa,$ff,$55
		ENDPATCH

		PATCH PATCH_draw_ammo0+1
		fcb $10	; ldd #$ff10
		ENDPATCH
		PATCH PATCH_draw_ammo1
		fcb $d4		; ldb #$d4
		ENDPATCH
		PATCH PATCH_draw_ammo2
		fcb $10		; ldb #$10
		ENDPATCH
		PATCH PATCH_draw_key01
		fcb $08,$a8	; ldd #$08a8
		ENDPATCH
		PATCH PATCH_draw_key11
		fcb $88		; lda #$88
		ENDPATCH

		PATCH snum_tiles_start
		fdb snum_tiles_ntsc1-5
		ENDPATCH

		ENDPATCHSET

; NTSC phase1 -> PAL patchset

ntsc1_pal_patchset

		PATCH colour_bg
		fcb $aa,$00,$55,$ff
		ENDPATCH

		PATCH vdg_mode
		fcb $e2		; lda #$e2 - VDG mode CG6, CSS0
		ENDPATCH

		PATCH logo
		fdb logo_pal
		ENDPATCH

		PATCH play_screen_dz
		fdb play_screen_pal_dz
		ENDPATCH

		PATCH text_screen_dz
		fdb text_screen_pal_dz
		ENDPATCH

		PATCH PATCH_draw_ammo0
		fdb $55ba	; ldd #$55ba
		ENDPATCH
		PATCH PATCH_draw_ammo1
		fcb $7e		; ldb #$7e
		ENDPATCH
		PATCH PATCH_draw_ammo2
		fcb $ba		; ldb #$ba
		ENDPATCH
		PATCH PATCH_draw_ammo3
		fcb $aa,$aa	; ldd #$aaaa
		ENDPATCH
		PATCH PATCH_draw_key0
		fdb $a656	; ldd #$a656
		ENDPATCH
		PATCH PATCH_draw_key1
		fcb $66		; lda #$66
		ENDPATCH
		PATCH PATCH_draw_key01
		fdb $a202	; ldd #$a202
		ENDPATCH
		PATCH PATCH_draw_key11
		fcb $22		; lda #$22
		ENDPATCH
		PATCH PATCH_undraw_key
		fcb $aa,$aa	; ldd #$aaaa
		ENDPATCH
		PATCH PATCH_draw_speed0
		fdb $a69a	; ldd #$a69a
		ENDPATCH
		PATCH PATCH_draw_speed1
		fcb $6a		; lda #$6a
		ENDPATCH
		PATCH PATCH_undraw_speed
		fcb $aa,$aa	; ldd #$aaaa
		ENDPATCH

		PATCH snum_tiles_start
		fdb snum_tiles_pal-5
		ENDPATCH

		ENDPATCHSET

		export cc3_cmp_rgb_patchset,cc3_rgb_cmp_patchset

; CoCo3 CMP->RGB patchset
cc3_cmp_rgb_patchset
		PATCH PATCH_cc3_c0c1
		fcb $12		; green (RGB)
		fcb $3e		; light yellow (RGB)
		ENDPATCH

		PATCH PATCH_cc3_c2c3
		fcb $00		; black (RGB)
		fcb $24		; red (RGB)
		ENDPATCH

		ENDPATCHSET

; CoCo3 RGB->CMP patchset
cc3_rgb_cmp_patchset
		PATCH PATCH_cc3_c0c1
		fcb $12		; green (CMP)
		fcb $34		; light yellow (CMP)
		ENDPATCH

		PATCH PATCH_cc3_c2c3
		fcb $00		; black (CMP)
		fcb $07		; red (CMP)
		ENDPATCH

		ENDPATCHSET

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		DATA

		include "fonts.s"
		include "logo-pal.s"
		include "logo-ntsc.s"
		include "snum-pal.s"
		include "snum-ntsc0.s"
		include "snum-ntsc1.s"

; ========================================================================

		export advance_video_mode
		export game_reset
		export tbl_kdsp_menu
		export start_game
		export toggle_p1
		export toggle_p2
		export toggle_p3
		export toggle_p4
		export toggle_skill
		export toggle_vmode
		export tbl_p1_keys
		export PATCH_text_p1_controls_ptr
		export PATCH_text_p2_controls_ptr
		export PATCH_p1ddrb

; Useful addresses for end of listing

		export CODE_start
		export CODE_end
		export DATA_start
		export DATA_end
		export INIT_end

CODE_start	equ CODE_start_
CODE_end	CODE
DATA_start	equ DATA_start_
DATA_end	DATA
BSS_start	equ BSS_start_
BSS_end		BSS
INIT_end	section "INIT"
