; Dunjunz - Games Master Edition
; Copyright (c) 2017-2020 Ciaran Anscomb
;
; See COPYING file for redistribution conditions.

; ========================================================================

; This file contains common macro definitions.

; ========================================================================

; Keyboard table entry macros to reduce possibility of error.

KEYDSP_ENTRY	macro
		fcb ~(1<<\1)		; column
		fcb 1<<\2		; row
		fdb \3			; dispatch
		endm

KEYTBL_ENTRY	macro
		fcb ~(1<<\1)		; column
		fcb 1<<\2		; row
		fcb \3			; action
		endm

; Head up a state machine jump table with this macro.

STATE_JTBL	macro
19824		equ *
		endm

; Use this macro to create symbols for specific states.

STATE_ID	macro
\1		equ *-19824B
		endm

; Patches that directly modify code and data to cope with different
; machine or display variants use a common format:

;   First entry:
;     patch byte count (0-255, 0 == 256 bytes)
;     base address
;     bytes...
;   Subsequent entries:
;     patch byte count (1-127, 0 == done, top bit set == 16-bit skip)
;     skip count
;       8-bit or 16-bit signed offset
;     bytes...

; Macro use:
;
; First use of PATCH macro (or after ENDPATCHSET) is special
; and embeds the absolute address of the first area to patch.

;               PATCH first_patch_address
;               fcb data,data,data
;               ENDPATCH
;
; Subsequent uses of PATCH macro embed relative addresses, using a single
; byte if possible to save space.

;               PATCH second_patch_address
;               fcb data,data,data
;               ENDPATCH

; ENDPATCHSET macro finishes a patchset, flagged with a zero size byte.

;               ENDPATCHSET

_PATCH_ADDR	set -1		; initially flag no patch in progress

PATCH		macro
	if (_PATCH_ADDR == -1)
		fcb 19826F-19825F
		fdb \1
	elsif (\1-_PATCH_ADDR) >= -128 && (\1-_PATCH_ADDR) < 128
		fcb 19826F-19825F
		fcb \1-_PATCH_ADDR
	else
		fcb $80|(19826F-19825F)
		fdb (\1-_PATCH_ADDR)
	endif
_PATCH_ADDR	set \1+19826F-19825F
19825
		endm

ENDPATCH	macro
19826
		endm

ENDPATCHSET	macro
		fcb 0
_PATCH_ADDR	set -1
		endm
