#!/bin/bash

PATH=/bin:/usr/bin
export PATH

. sheetfuncs.sh

test -x "$TILE2BIN" || { echo "$TILE2BIN: not executable" >&2; exit 1; }

sheet="tiles.png"

suffix="$1"
shift

if test "$suffix" = "ntsc0"; then
	map="-m 1,3,0,2"
elif test "$suffix" = "ntsc1"; then
	map="-m 2,3,0,1"
fi

W=12
H=9

# Updated for GMC - don't bother emitting labels, as we need to generate
# every shifted set now.

label "tiles_$suffix"

echo
echo "; door graphics"

bitmap_a "$sheet" tile_door1_v  $((W*5)) $((H*11)) $map	#0
bitmap_a "$sheet" tile_door1_h  $((W*6)) $((H*11)) $map	#1
bitmap_a "$sheet" tile_door_v   $((W*5)) $((H*4)) $map	#2
bitmap_a "$sheet" tile_door_h   $((W*6)) $((H*4)) $map	#3

echo
echo "; misc items"

bitmap_a "$sheet" tile_floor    $((W*4)) $((H*0)) $map	#4
bitmap_a "$sheet" tile_exit     $((W*5)) $((H*0)) $map	#5
bitmap_a "$sheet" tile_trapdoor $((W*6)) $((H*0)) $map	#6
bitmap_a "$sheet" tile_food     $((W*4)) $((H*3)) $map	#7

bitmap_a "$sheet" tile_tport    $((W*7)) $((H*0)) $map	#8
bitmap_a "$sheet" tile_armour   $((W*7)) $((H*2)) $map	#9
bitmap_a "$sheet" tile_potion   $((W*5)) $((H*3)) $map	#10
bitmap_a "$sheet" tile_weapon   $((W*6)) $((H*2)) $map	#11

bitmap_a "$sheet" tile_cross    $((W*6)) $((H*3)) $map	#12
bitmap_a "$sheet" tile_speed    $((W*7)) $((H*3)) $map	#13
bitmap_a "$sheet" tile_drainer0 $((W*4)) $((H*1)) $map	#14
bitmap_a "$sheet" tile_money0   $((W*6)) $((H*1)) $map	#15

bitmap_a "$sheet" tile_power0   $((W*4)) $((H*2)) $map	#16
bitmap_a "$sheet" tile_drainer1 $((W*5)) $((H*1)) $map	#17 XXX alignment

bitmap_a "$sheet" tile_key      $((W*7)) $((H*4)) $map	#18
bitmap_a "$sheet" tile_key1     $((W*7)) $((H*11)) $map	#19

echo
echo "; player/monster graphics"

bitmap_01_a "$sheet" tile_p1_up    $((W*0)) $((H*0)) $map	#20-21
bitmap_01_a "$sheet" tile_p1_down  $((W*0)) $((H*0)) -X -Y $map	#22-23
bitmap_01_a "$sheet" tile_p1_left  $((W*2)) $((H*0)) $map	#24-25
bitmap_01_a "$sheet" tile_p1_right $((W*2)) $((H*0)) -X -Y $map	#26-27

bitmap_a "$sheet" tile_arrow_up    $((W*0)) $((H*5)) $map	#28
bitmap_a "$sheet" tile_arrow_down  $((W*0)) $((H*5)) -X -Y $map	#29
bitmap_a "$sheet" tile_arrow_left  $((W*1)) $((H*5)) $map	#30
bitmap_a "$sheet" tile_arrow_right $((W*1)) $((H*5)) -X -Y $map	#31

bitmap_a "$sheet" tile_p1_exit4 $((W*0)) $((H*9)) $map		#32
bitmap_a "$sheet" tile_p1_exit5 $((W*1)) $((H*9)) $map		#33
bitmap_a "$sheet" tile_p1_exit6 $((W*0)) $((H*9)) -X -Y $map	#34
bitmap_a "$sheet" tile_p1_exit7 $((W*1)) $((H*9)) -X -Y $map	#35

bitmap_01_a "$sheet" tile_p2_up    $((W*0)) $((H*1)) $map	#36-37
bitmap_01_a "$sheet" tile_p2_down  $((W*0)) $((H*1)) -X -Y $map	#38-39
bitmap_01_a "$sheet" tile_p2_left  $((W*2)) $((H*1)) $map	#40-41
bitmap_01_a "$sheet" tile_p2_right $((W*2)) $((H*1)) -X -Y $map	#42-43

bitmap_a "$sheet" tile_fball_up    $((W*2)) $((H*5)) $map	#44
bitmap_a "$sheet" tile_fball_down  $((W*2)) $((H*5)) -X -Y $map	#45
bitmap_a "$sheet" tile_fball_left  $((W*3)) $((H*5)) $map	#46
bitmap_a "$sheet" tile_fball_right $((W*3)) $((H*5)) -X -Y $map	#47

bitmap_a "$sheet" tile_p2_exit4 $((W*2)) $((H*9)) $map		#48
bitmap_a "$sheet" tile_p2_exit5 $((W*3)) $((H*9)) $map		#49
bitmap_a "$sheet" tile_p2_exit6 $((W*2)) $((H*9)) -X -Y $map	#50
bitmap_a "$sheet" tile_p2_exit7 $((W*3)) $((H*9)) -X -Y $map	#51

bitmap_01_a "$sheet" tile_p3_up    $((W*0)) $((H*2)) $map	#52-53
bitmap_01_a "$sheet" tile_p3_down  $((W*0)) $((H*2)) -X -Y $map	#54-55
bitmap_01_a "$sheet" tile_p3_left  $((W*2)) $((H*2)) $map	#56-57
bitmap_01_a "$sheet" tile_p3_right $((W*2)) $((H*2)) -X -Y $map	#58-59

bitmap_a "$sheet" tile_axe_up      $((W*0)) $((H*6)) $map	#60
bitmap_a "$sheet" tile_axe_down    $((W*0)) $((H*6)) -X -Y $map	#61
bitmap_a "$sheet" tile_axe_left    $((W*1)) $((H*6)) $map	#62
bitmap_a "$sheet" tile_axe_right   $((W*1)) $((H*6)) -X -Y $map	#63

bitmap_a "$sheet" tile_p3_exit4 $((W*0)) $((H*10)) $map		#64
bitmap_a "$sheet" tile_p3_exit5 $((W*1)) $((H*10)) $map		#65
bitmap_a "$sheet" tile_p3_exit6 $((W*0)) $((H*10)) -X -Y $map	#66
bitmap_a "$sheet" tile_p3_exit7 $((W*1)) $((H*10)) -X -Y $map	#67

bitmap_01_a "$sheet" tile_p4_up    $((W*0)) $((H*3)) $map	#68-69
bitmap_01_a "$sheet" tile_p4_down  $((W*0)) $((H*3)) -X -Y $map	#70-71
bitmap_01_a "$sheet" tile_p4_left  $((W*2)) $((H*3)) $map	#72-73
bitmap_01_a "$sheet" tile_p4_right $((W*2)) $((H*3)) -X -Y $map	#74-74

bitmap_a "$sheet" tile_sword_up    $((W*2)) $((H*6)) $map	#76
bitmap_a "$sheet" tile_sword_down  $((W*2)) $((H*6)) -X -Y $map	#77
bitmap_a "$sheet" tile_sword_left  $((W*3)) $((H*6)) $map	#78
bitmap_a "$sheet" tile_sword_right $((W*3)) $((H*6)) -X -Y $map	#79

bitmap_a "$sheet" tile_p4_exit4 $((W*2)) $((H*10)) $map		#80
bitmap_a "$sheet" tile_p4_exit5 $((W*3)) $((H*10)) $map		#81
bitmap_a "$sheet" tile_p4_exit6 $((W*2)) $((H*10)) -X -Y $map	#82
bitmap_a "$sheet" tile_p4_exit7 $((W*3)) $((H*10)) -X -Y $map	#83

echo
echo "; exit animations"

bitmap_a "$sheet" tile_exit8 $((W*0)) $((H*11)) $map		#84
bitmap_a "$sheet" tile_exit9 $((W*1)) $((H*11)) $map		#85
bitmap_a "$sheet" tile_exit10 $((W*0)) $((H*11)) -X -Y $map	#86
bitmap_a "$sheet" tile_exit11 $((W*1)) $((H*11)) -X -Y $map	#87

bitmap_01_a "$sheet" tile_monster_up    $((W*0)) $((H*4)) $map		#88-89
bitmap_01_a "$sheet" tile_monster_down  $((W*0)) $((H*4)) -X -Y $map	#90-91
bitmap_01_a "$sheet" tile_monster_left  $((W*2)) $((H*4)) $map		#92-93
bitmap_01_a "$sheet" tile_monster_right $((W*2)) $((H*4)) -X -Y $map	#94-95

echo
echo "; bones - death animation"

bitmap_a "$sheet" tile_bones0      $((W*0)) $((H*7)) $map	#96
bitmap_a "$sheet" tile_bones1      $((W*1)) $((H*7)) $map	#97
bitmap_a "$sheet" tile_bones2      $((W*2)) $((H*7)) $map	#97
bitmap_a "$sheet" tile_bones3      $((W*3)) $((H*7)) $map	#99

bitmap_a "$sheet" tile_bones4      $((W*0)) $((H*8)) $map	#100
bitmap_a "$sheet" tile_bones5      $((W*1)) $((H*8)) $map	#101
bitmap_a "$sheet" tile_bones6      $((W*2)) $((H*8)) $map	#102
bitmap_a "$sheet" tile_bones7      $((W*3)) $((H*8)) $map	#103

echo
echo "; alternative animation frames"
echo "; note - tile_drainer1" appears already as #17

bitmap_a "$sheet" tile_money1   $((W*7)) $((H*1)) $map	#104
bitmap_a "$sheet" tile_power1   $((W*5)) $((H*2)) $map  #105

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

echo
echo "; door graphics"

bitmap_b "$sheet" tile_door1_v  $((W*5)) $((H*11)) $map
bitmap_b "$sheet" tile_door1_h  $((W*6)) $((H*11)) $map
bitmap_b "$sheet" tile_door_v   $((W*5)) $((H*4)) $map
bitmap_b "$sheet" tile_door_h   $((W*6)) $((H*4)) $map

echo
echo "; misc items"

bitmap_b "$sheet" tile_floor    $((W*4)) $((H*0)) $map
bitmap_b "$sheet" tile_exit     $((W*5)) $((H*0)) $map
bitmap_b "$sheet" tile_trapdoor $((W*6)) $((H*0)) $map
bitmap_b "$sheet" tile_food     $((W*4)) $((H*3)) $map

bitmap_b "$sheet" tile_tport    $((W*7)) $((H*0)) $map
bitmap_b "$sheet" tile_armour   $((W*7)) $((H*2)) $map
bitmap_b "$sheet" tile_potion   $((W*5)) $((H*3)) $map
bitmap_b "$sheet" tile_weapon   $((W*6)) $((H*2)) $map

bitmap_b "$sheet" tile_cross    $((W*6)) $((H*3)) $map
bitmap_b "$sheet" tile_speed    $((W*7)) $((H*3)) $map
bitmap_b "$sheet" tile_drainer0 $((W*4)) $((H*1)) $map
bitmap_b "$sheet" tile_money0   $((W*6)) $((H*1)) $map

bitmap_b "$sheet" tile_power0   $((W*4)) $((H*2)) $map
bitmap_b "$sheet" tile_drainer1 $((W*5)) $((H*1)) $map	# XXX alignment

bitmap_b "$sheet" tile_key      $((W*7)) $((H*4)) $map
bitmap_b "$sheet" tile_key1     $((W*7)) $((H*11)) $map

echo
echo "; player/monster graphics"

bitmap_01_b "$sheet" tile_p1_up    $((W*0)) $((H*0)) $map
bitmap_01_b "$sheet" tile_p1_down  $((W*0)) $((H*0)) -X -Y $map
bitmap_01_b "$sheet" tile_p1_left  $((W*2)) $((H*0)) $map
bitmap_01_b "$sheet" tile_p1_right $((W*2)) $((H*0)) -X -Y $map

bitmap_b "$sheet" tile_arrow_up    $((W*0)) $((H*5)) $map
bitmap_b "$sheet" tile_arrow_down  $((W*0)) $((H*5)) -X -Y $map
bitmap_b "$sheet" tile_arrow_left  $((W*1)) $((H*5)) $map
bitmap_b "$sheet" tile_arrow_right $((W*1)) $((H*5)) -X -Y $map

bitmap_b "$sheet" tile_p1_exit4 $((W*0)) $((H*9)) $map
bitmap_b "$sheet" tile_p1_exit5 $((W*1)) $((H*9)) $map
bitmap_b "$sheet" tile_p1_exit6 $((W*0)) $((H*9)) -X -Y $map
bitmap_b "$sheet" tile_p1_exit7 $((W*1)) $((H*9)) -X -Y $map

bitmap_01_b "$sheet" tile_p2_up    $((W*0)) $((H*1)) $map
bitmap_01_b "$sheet" tile_p2_down  $((W*0)) $((H*1)) -X -Y $map
bitmap_01_b "$sheet" tile_p2_left  $((W*2)) $((H*1)) $map
bitmap_01_b "$sheet" tile_p2_right $((W*2)) $((H*1)) -X -Y $map

bitmap_b "$sheet" tile_fball_up    $((W*2)) $((H*5)) $map
bitmap_b "$sheet" tile_fball_down  $((W*2)) $((H*5)) -X -Y $map
bitmap_b "$sheet" tile_fball_left  $((W*3)) $((H*5)) $map
bitmap_b "$sheet" tile_fball_right $((W*3)) $((H*5)) -X -Y $map

bitmap_b "$sheet" tile_p2_exit4 $((W*2)) $((H*9)) $map
bitmap_b "$sheet" tile_p2_exit5 $((W*3)) $((H*9)) $map
bitmap_b "$sheet" tile_p2_exit6 $((W*2)) $((H*9)) -X -Y $map
bitmap_b "$sheet" tile_p2_exit7 $((W*3)) $((H*9)) -X -Y $map

bitmap_01_b "$sheet" tile_p3_up    $((W*0)) $((H*2)) $map
bitmap_01_b "$sheet" tile_p3_down  $((W*0)) $((H*2)) -X -Y $map
bitmap_01_b "$sheet" tile_p3_left  $((W*2)) $((H*2)) $map
bitmap_01_b "$sheet" tile_p3_right $((W*2)) $((H*2)) -X -Y $map

bitmap_b "$sheet" tile_axe_up      $((W*0)) $((H*6)) $map
bitmap_b "$sheet" tile_axe_down    $((W*0)) $((H*6)) -X -Y $map
bitmap_b "$sheet" tile_axe_left    $((W*1)) $((H*6)) $map
bitmap_b "$sheet" tile_axe_right   $((W*1)) $((H*6)) -X -Y $map

bitmap_b "$sheet" tile_p3_exit4 $((W*0)) $((H*10)) $map
bitmap_b "$sheet" tile_p3_exit5 $((W*1)) $((H*10)) $map
bitmap_b "$sheet" tile_p3_exit6 $((W*0)) $((H*10)) -X -Y $map
bitmap_b "$sheet" tile_p3_exit7 $((W*1)) $((H*10)) -X -Y $map

bitmap_01_b "$sheet" tile_p4_up    $((W*0)) $((H*3)) $map
bitmap_01_b "$sheet" tile_p4_down  $((W*0)) $((H*3)) -X -Y $map
bitmap_01_b "$sheet" tile_p4_left  $((W*2)) $((H*3)) $map
bitmap_01_b "$sheet" tile_p4_right $((W*2)) $((H*3)) -X -Y $map

bitmap_b "$sheet" tile_sword_up    $((W*2)) $((H*6)) $map
bitmap_b "$sheet" tile_sword_down  $((W*2)) $((H*6)) -X -Y $map
bitmap_b "$sheet" tile_sword_left  $((W*3)) $((H*6)) $map
bitmap_b "$sheet" tile_sword_right $((W*3)) $((H*6)) -X -Y $map

bitmap_b "$sheet" tile_p4_exit4 $((W*2)) $((H*10)) $map
bitmap_b "$sheet" tile_p4_exit5 $((W*3)) $((H*10)) $map
bitmap_b "$sheet" tile_p4_exit6 $((W*2)) $((H*10)) -X -Y $map
bitmap_b "$sheet" tile_p4_exit7 $((W*3)) $((H*10)) -X -Y $map

echo
echo "; exit animations"

bitmap_b "$sheet" tile_exit8 $((W*0)) $((H*11)) $map
bitmap_b "$sheet" tile_exit9 $((W*1)) $((H*11)) $map
bitmap_b "$sheet" tile_exit10 $((W*0)) $((H*11)) -X -Y $map
bitmap_b "$sheet" tile_exit11 $((W*1)) $((H*11)) -X -Y $map

bitmap_01_b "$sheet" tile_monster_up    $((W*0)) $((H*4)) $map
bitmap_01_b "$sheet" tile_monster_down  $((W*0)) $((H*4)) -X -Y $map
bitmap_01_b "$sheet" tile_monster_left  $((W*2)) $((H*4)) $map
bitmap_01_b "$sheet" tile_monster_right $((W*2)) $((H*4)) -X -Y $map

echo
echo "; bones - death animation"

bitmap_b "$sheet" tile_bones0      $((W*0)) $((H*7)) $map
bitmap_b "$sheet" tile_bones1      $((W*1)) $((H*7)) $map
bitmap_b "$sheet" tile_bones2      $((W*2)) $((H*7)) $map
bitmap_b "$sheet" tile_bones3      $((W*3)) $((H*7)) $map

bitmap_b "$sheet" tile_bones4      $((W*0)) $((H*8)) $map
bitmap_b "$sheet" tile_bones5      $((W*1)) $((H*8)) $map
bitmap_b "$sheet" tile_bones6      $((W*2)) $((H*8)) $map
bitmap_b "$sheet" tile_bones7      $((W*3)) $((H*8)) $map

echo
echo "; alternative animation frames"

bitmap_b "$sheet" tile_money1   $((W*7)) $((H*1)) $map
bitmap_b "$sheet" tile_power1   $((W*5)) $((H*2)) $map

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

# Tick, cross, chalice graphics only ever appear in one position, and so
# shifted versions are not required.

echo
echo "; yes/no (tick/cross)"

bitmap_b "$sheet" tile_yes $((W*0)) $((H*12)) $map	#0
bitmap_b "$sheet" tile_no  $((W*1)) $((H*12)) $map	#1

echo
echo "; chalice"

bitmap_a "$sheet" tile_chalice_nw0 $((W*2)) $((H*11)) $map	#2
bitmap_b "$sheet" tile_chalice_ne0 $((W*2)) $((H*11)) -X $map
bitmap_a "$sheet" tile_chalice_sw0 $((W*2)) $((H*11)) -Y $map
bitmap_b "$sheet" tile_chalice_se0 $((W*2)) $((H*11)) -X -Y $map

bitmap_a "$sheet" tile_chalice_nw1 $((W*3)) $((H*11)) $map	#6
bitmap_b "$sheet" tile_chalice_ne1 $((W*3)) $((H*11)) -X $map
bitmap_a "$sheet" tile_chalice_sw1 $((W*3)) $((H*11)) -Y $map
bitmap_b "$sheet" tile_chalice_se1 $((W*3)) $((H*11)) -X -Y $map

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

echo
echo "; stones"

bitmap_a "$sheet" tile_stone01 $((W*4)) $((H*5)) $map
bitmap_a "$sheet" tile_stone02 $((W*5)) $((H*5)) $map
bitmap_a "$sheet" tile_stone03 $((W*6)) $((H*5)) $map
bitmap_a "$sheet" tile_stone04 $((W*7)) $((H*5)) $map

bitmap_a "$sheet" tile_stone05 $((W*4)) $((H*6)) $map
bitmap_a "$sheet" tile_stone06 $((W*5)) $((H*6)) $map
bitmap_a "$sheet" tile_stone07 $((W*6)) $((H*6)) $map
bitmap_a "$sheet" tile_stone08 $((W*7)) $((H*6)) $map

bitmap_a "$sheet" tile_stone09 $((W*4)) $((H*7)) $map
bitmap_a "$sheet" tile_stone10 $((W*5)) $((H*7)) $map
bitmap_a "$sheet" tile_stone11 $((W*6)) $((H*7)) $map
bitmap_a "$sheet" tile_stone12 $((W*7)) $((H*7)) $map

bitmap_a "$sheet" tile_stone13 $((W*4)) $((H*8)) $map
bitmap_a "$sheet" tile_stone14 $((W*5)) $((H*8)) $map
bitmap_a "$sheet" tile_stone15 $((W*6)) $((H*8)) $map
bitmap_a "$sheet" tile_stone16 $((W*7)) $((H*8)) $map

bitmap_a "$sheet" tile_stone17 $((W*4)) $((H*9)) $map
bitmap_a "$sheet" tile_stone18 $((W*5)) $((H*9)) $map
bitmap_a "$sheet" tile_stone19 $((W*6)) $((H*9)) $map
bitmap_a "$sheet" tile_stone20 $((W*7)) $((H*9)) $map

bitmap_a "$sheet" tile_stone21 $((W*4)) $((H*10)) $map
bitmap_a "$sheet" tile_stone22 $((W*5)) $((H*10)) $map
bitmap_a "$sheet" tile_stone23 $((W*6)) $((H*10)) $map
bitmap_a "$sheet" tile_stone24 $((W*7)) $((H*10)) $map

bitmap_a "$sheet" tile_stone25 $((W*4)) $((H*11)) $map

bitmap_b "$sheet" tile_stone01 $((W*4)) $((H*5)) $map
bitmap_b "$sheet" tile_stone02 $((W*5)) $((H*5)) $map
bitmap_b "$sheet" tile_stone03 $((W*6)) $((H*5)) $map
bitmap_b "$sheet" tile_stone04 $((W*7)) $((H*5)) $map

bitmap_b "$sheet" tile_stone05 $((W*4)) $((H*6)) $map
bitmap_b "$sheet" tile_stone06 $((W*5)) $((H*6)) $map
bitmap_b "$sheet" tile_stone07 $((W*6)) $((H*6)) $map
bitmap_b "$sheet" tile_stone08 $((W*7)) $((H*6)) $map

bitmap_b "$sheet" tile_stone09 $((W*4)) $((H*7)) $map
bitmap_b "$sheet" tile_stone10 $((W*5)) $((H*7)) $map
bitmap_b "$sheet" tile_stone11 $((W*6)) $((H*7)) $map
bitmap_b "$sheet" tile_stone12 $((W*7)) $((H*7)) $map

bitmap_b "$sheet" tile_stone13 $((W*4)) $((H*8)) $map
bitmap_b "$sheet" tile_stone14 $((W*5)) $((H*8)) $map
bitmap_b "$sheet" tile_stone15 $((W*6)) $((H*8)) $map
bitmap_b "$sheet" tile_stone16 $((W*7)) $((H*8)) $map

bitmap_b "$sheet" tile_stone17 $((W*4)) $((H*9)) $map
bitmap_b "$sheet" tile_stone18 $((W*5)) $((H*9)) $map
bitmap_b "$sheet" tile_stone19 $((W*6)) $((H*9)) $map
bitmap_b "$sheet" tile_stone20 $((W*7)) $((H*9)) $map

bitmap_b "$sheet" tile_stone21 $((W*4)) $((H*10)) $map
bitmap_b "$sheet" tile_stone22 $((W*5)) $((H*10)) $map
bitmap_b "$sheet" tile_stone23 $((W*6)) $((H*10)) $map
bitmap_b "$sheet" tile_stone24 $((W*7)) $((H*10)) $map

bitmap_b "$sheet" tile_stone25 $((W*4)) $((H*11)) $map

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

